"""Command line interface to e(BE:L)."""

import sys
import logging

import click

import ebel.database
from ebel import Bel, web
from ebel.config import user_config_setup, set_configuration
from ebel.manager.orientdb.constants import DRUGBANK
from ebel.validate import validate_bel_file, repair_bel_file

logger = logging.getLogger(__name__)


@click.group(help="BEL test framework Command Line Utilities on {}".format(sys.executable))
@click.version_option()
def main():
    """Entry method."""
    pass


# TODO(@ceb): Implement by default "keep database" and new database as option #prio1
# TODO(@ceb): SQLAlchemy connection string #prio1
@main.command()
@click.argument('bel_script_path')
@click.option('-l', '--line_by_line', is_flag=True, default=False, help='check script line by line')
@click.option('-r', '--reports', default=None,
              help='path(s) to report file(s) seperated by comma with suffix (.md, .txt, .csv, .tsv, .json, .html)')
@click.option('-v', '--bel_version', default="2_1", help='stores a report in a file')
@click.option('-t', '--tree', is_flag=True, default=False, help='shows tree')
# TODO: Implement new Cytoscape export file
# @click.option('-c', '--cytoscape', is_flag=True, default=False, help='creates cytoscape file')
@click.option('-s', '--sqlalchemy_connection_str', default=None, help="SQLAlchmy connection string")
@click.option('-j', '--json_file', is_flag=True, help="Create json file")
@click.option('-f', '--force_json', is_flag=True, default=False, help="Force the creation of a JSON file")
def validate(bel_script_path: str, line_by_line: bool, reports: str,
             bel_version: str, tree: bool, sqlalchemy_connection_str: str,
             json_file: bool, force_json: bool):
    """Validate a BEL file using the defined grammar."""
    validate_bel_file(bel_script_path=bel_script_path, line_by_line=line_by_line, reports=reports,
                      bel_version=bel_version, tree=tree, sqlalchemy_connection_str=sqlalchemy_connection_str,
                      json_file=json_file, force_json=force_json)


@main.command()
@click.argument('bel_script_path')
@click.option('-n', '--new_file_path', default=None,
              help='Path to write repaired file to. If none passed, will overwrite original file.')
def repair(bel_script_path: str, new_file_path: str):
    """Repair the BEL file for common delimiters and line separations."""
    repair_bel_file(bel_script_path, new_file_path)


@main.command()
@click.argument('json_file_path')
@click.option('-e', '--extend', is_flag=True, default=True, help='Flag to disable "extension" during import')
@click.option('-g', '--p2g', is_flag=True, default=True, help='Flag to disable "protein2gene" during import')
@click.option('-s', '--skip_drugbank', is_flag=True, default=False, help='Flag to disable DrugBank')
@click.option('-i', '--include_subfolders', is_flag=True, default=False, help='Flag to enable directory walking')
@click.option('--drugbank_user', help='Valid username for DrugBank')
@click.option('--drugbank_password', help='Valid username for DrugBank')
@click.option('-n', '--odb_name', help='OrientDB database name')
@click.option('-u', '--odb_user', help='OrientDB user (with admin rights)')
@click.option('-p', '--odb_password', help='OrientDB user (with admin rights) password')
@click.option('-h', '--odb_server', help='OrientDB server name or URI')
@click.option('-o', '--odb_port', help='OrientDB server port')
@click.option('--odb_user_reader', help=' OrientDB user with only read rights')
@click.option('--odb_user_reader_password', help='OrientDB user with only read rights password')
@click.option('--odb_root_password', help='OrientDB root user password (only during database setup)')
@click.option('--kegg_species', help='KEGG species')
@click.option('--sqlalchemy_connection_string', help='schema is user:password@server/database')
@click.option('--snp_related_traits', help='key of SNP related traits in GWAS catalog')
@click.option('--drugbank_user', help='Drugbank user')
@click.option('--drugbank_password', help='DrugBank password')
def import_json(json_file_path: str,
                extend: bool,
                p2g: bool,
                skip_drugbank: bool,
                drugbank_user: str,
                drugbank_password: str,
                include_subfolders: bool,
                odb_name: str,
                odb_user: str,
                odb_password: str,
                odb_server: str,
                odb_port: str,
                odb_user_reader: str,
                odb_user_reader_password: str,
                odb_root_password: str,
                kegg_species: str,
                sqlalchemy_connection_string: str,
                snp_related_traits: str):
    """Import JSON into OrientDB.

    Parameters
    ----------
    json_file_path : str
        Path to a BEL JSON file, list of BEL JSON files (separated by ","), or a directory containing BEL JSON files.
    extend : bool
        Boolean flag to disable network extension by e(BE:L)
    p2g : bool
        Boolean flag to disable adding genes and RNA to proteins in network
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    include_subfolders: bool (optional)
        Boolean flag to enable walking through subdirectories during import.
    odb_name: str
        OrientDB database name
    odb_user:
        OrientDB user (with admin rights)
    odb_password:
        OrientDB user (with admin rights) password
    odb_server:
        OrientDB server name or URI
    odb_port:
        OrientDB server port
    odb_user_reader:
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password:
        OrientDB root user password (only first database setup needed)
    kegg_species:
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits:
        SNP related traits

    Returns
    -------
    type
        returns True if imported.
    """
    # if one of the parameters is not None it will overwrite the default values from the configfile
    set_configuration(name=odb_name,
                      user=odb_user,
                      password=odb_password,
                      server=odb_server,
                      port=odb_port,
                      user_reader=odb_user_reader,
                      user_reader_password=odb_user_reader_password,
                      root_password=odb_root_password,
                      kegg_species=kegg_species,
                      sqlalchemy_connection_string=sqlalchemy_connection_string,
                      snp_related_traits=snp_related_traits,
                      drugbank_user=drugbank_user,
                      drugbank_password=drugbank_password)

    bel = Bel()
    if "," in json_file_path:
        json_file_path = json_file_path.split()

    bel.import_json(input_path=json_file_path,
                    extend_graph=extend,
                    update_from_protein2gene=p2g,
                    skip_drugbank=skip_drugbank,
                    drugbank_user=drugbank_user,
                    drugbank_password=drugbank_password,
                    include_subfolders=include_subfolders)


@main.command()
@click.option('-s', '--skip', default=[], help='Comma-separated list of databases to skip during enrichment')
@click.option('-i', '--include', default=[], help='Comma-separated list of databases to include during enrichment')
@click.option('--skip_drugbank', is_flag=True, default=False, help='Flag to disable DrugBank')
@click.option('--drugbank_user', default=None, help='Valid username for DrugBank')
@click.option('--drugbank_password', default=None, help='Valid username for DrugBank')
@click.option('-n', '--odb_name', default=None, help='OrientDB database name')
@click.option('-u', '--odb_user', default=None, help='OrientDB user (with admin rights)')
@click.option('-p', '--odb_password', default=None, help='OrientDB user (with admin rights) password')
@click.option('-h', '--odb_server', default=None, help='OrientDB server name or URI')
@click.option('-o', '--odb_port', default=None, help='OrientDB server port')
@click.option('--odb_user_reader', default=None, help=' OrientDB user with only read rights')
@click.option('--odb_user_reader_password', default=None, help='OrientDB user with only read rights password')
@click.option('--odb_root_password', default=None, help='OrientDB root user password (only during database setup)')
@click.option('--kegg_species', default='hsa,rno,mmu', help='KEGG species')
@click.option('--sqlalchemy_connection_string', help='schema is user:password@server/database')
@click.option('--snp_related_traits', help='key of SNP related traits in GWAS catalog and ClinVar')
def enrich(skip: str,
           include: str,
           skip_drugbank: bool,
           drugbank_user: str,
           drugbank_password: str,
           odb_name: str,
           odb_user: str,
           odb_password: str,
           odb_server: str,
           odb_port: str,
           odb_user_reader: str,
           odb_user_reader_password: str,
           odb_root_password: str,
           kegg_species: str,
           sqlalchemy_connection_string: str,
           snp_related_traits: str):
    """Trigger the enrichment step for a database.

    Parameters
    ----------
    skip : str
        Comma-separated list of databases to skip during enrichment.
    include : str
        Comma-separated list of databases to include during enrichment.
    skip_drugbank : bool
        Boolean flag to disable DrugBank enrichment. Useful if you do not have valid DrugBank credentials
    drugbank_user: str (optional)
        DrugBank user name.
    drugbank_password: str (optional)
        DrugBank password.
    odb_name: str
        OrientDB database name
    odb_user: str
        OrientDB user (with admin rights)
    odb_password: str
        OrientDB user (with admin rights) password
    odb_server: str
        OrientDB server name or URI
    odb_port: int
        OrientDB server port
    odb_user_reader: str
        OrientDB user with only read rights
    odb_user_reader_password:
        OrientDB user with only read rights password
    odb_root_password: str
        OrientDB root user password (only first database setup needed)
    kegg_species: str
        KEGG species 3-4 letter code https://www.genome.jp/kegg/catalog/org_list4.html
    sqlalchemy_connection_string:
        SQL Alchemy connection string schema: user:passwd@server/database
    snp_related_traits: str
        SNP related traits
    """
    set_configuration(name=odb_name,
                      user=odb_user,
                      password=odb_password,
                      server=odb_server,
                      port=odb_port,
                      user_reader=odb_user_reader,
                      user_reader_password=odb_user_reader_password,
                      root_password=odb_root_password,
                      kegg_species=kegg_species,
                      sqlalchemy_connection_string=sqlalchemy_connection_string,
                      snp_related_traits=snp_related_traits,
                      drugbank_user=drugbank_user,
                      drugbank_password=drugbank_password)

    bel = Bel()

    if not skip_drugbank:
        bel.drugbank.get_user_passwd(drugbank_user=drugbank_user, drugbank_password=drugbank_password)

    skip = skip.split(",") if isinstance(skip, str) else skip
    include = include.split(",") if isinstance(include, str) else include

    if skip_drugbank and DRUGBANK not in skip:
        skip.append(DRUGBANK)

    bel.enrich_network(skip=skip, include=include)


@main.command()
@click.argument('connection')
def set_connection(connection):
    """Set the SQLAlchemy connection string.

    :param connection: An existing RDBMS connection.
    """
    ebel.database.set_connection(connection)


@main.command()
@click.option('-h', '--host', default='localhost', help="MySQL server")
@click.option('-u', '--user', default='ebel_user', help="MySQL username")
@click.option('-p', '--password', default='ebel_passwd', help="MySQL password")
@click.option('-d', '--database', default='ebel', help="MySQL database name")
@click.option('-i', '--interactive', is_flag=True, default=False, help="Enable interactive mode")
def set_mysql(host: str, user: str, password: str, database: str, interactive: bool):
    """Set the SQLAlchemy connection string with MySQL settings."""
    if interactive:
        host, user, password, db = ebel.database.set_mysql_interactive()

    ebel.database.set_mysql_connection(
        host=host,
        user=user,
        password=password,
        db=database
    )


@main.command()
def settings():
    """Interactive method to create a configuration file."""
    user_config_setup()


@main.command()
@click.option('-h', '--host', default='0.0.0.0', help='Server or host name')
@click.option('-p', '--port', default='5000', help='server port [5000]')
@click.option('-d', '--debug_mode', is_flag=True, default=False, help='debug mode')
@click.option('-o', '--open_browser', is_flag=True, default=False, help='open browser')
def serve(host, port, debug_mode, open_browser):
    """Start the API RESTful server."""
    web.app.run(host=host, port=port, debug_mode=debug_mode, open_browser=open_browser)


if __name__ == '__main__':
    main()



"""ClinVar API methods."""

from flask import request

from ebel import Bel
from math import ceil
from ebel.web.api import RDBMS
from ebel.manager.rdbms.models import clinvar
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_query_result, _get_terms_from_model_starts_with, \
    _get_pagination


def get_clinvar():
    """Get generic ClinVar entry."""
    return _get_data(clinvar.Clinvar)


def get_clinvar_simple():
    """Generic Clinvar result query."""
    cc = clinvar.Clinvar
    cp = clinvar.ClinvarPhenotype
    paras = {k: v for k, v in request.args.items() if
             k in ['gene_id', 'gene_symbol', 'hgnc_id', 'allele_id', 'assembly', 'rs_db_snp']}
    query = RDBMS.get_session().query(cc.id, cc.hgnc_id, cc.allele_id, cc.gene_symbol, cc.assembly, cc.rs_db_snp)\
        .filter_by(**paras).join(clinvar.clinvar__clinvar_phenotype).join(cp)

    phenotype = request.args.get('phenotype')
    if phenotype:
        query = query.filter_by(phenotype='Alzheimer disease')

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_phenotype_starts_with():
    """Get Clinvar results by fuzzy phenotype search."""
    return _get_terms_from_model_starts_with('phenotype', clinvar.ClinvarPhenotype.phenotype)


def get_by_other_identifier():
    """Get Clinvar results by other identifier."""
    cc = clinvar.Clinvar
    co = clinvar.ClinvarOtherIdentifier

    db = request.args.get('db')
    identifier = request.args.get('identifier')
    query = RDBMS.get_session().query(co.db,
                                      co.identifier,
                                      cc.id,
                                      cc.hgnc_id,
                                      cc.allele_id,
                                      cc.gene_symbol,
                                      cc.assembly,
                                      cc.rs_db_snp).join(co).filter_by(db=db).filter(co.identifier.like(identifier))

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_by_medgen():
    """Get Clinvar results by MedGen identifier."""
    cc = clinvar.Clinvar
    cm = clinvar.ClinvarPhenotypeMedgen

    # db = request.args.get('db')
    identifier = request.args.get('identifier')
    query = RDBMS.get_session().query(cm.identifier,
                                      cc.id,
                                      cc.hgnc_id,
                                      cc.allele_id,
                                      cc.gene_symbol,
                                      cc.assembly,
                                      cc.rs_db_snp).join(cm).filter_by(identifier=identifier)

    return _get_paginated_query_result(query.distinct(), return_dict=True)


def get_ebel_relation():
    """Get ClinVar related e(BE:L) relations."""
    b = Bel()
    p = _get_pagination()
    req = dict(request.args.copy())
    wheres = []
    if req.get('rs_number'):
        wheres.append(f"in.rs_number='{req['rs_number']}'")
    relation_type = req.get('relation_type')
    allowed_relations = ['has_snp_clinvar',
                         'has_mapped_snp_cv',
                         'has_downstream_snp_cv',
                         'has_upstream_snp_cv']
    relation_type = relation_type if relation_type in allowed_relations else allowed_relations[0]
    wheres += [f'out.{k} = "{v}"' for k, v in req.items() if k in ['namespace', 'name']]
    wheres += [f'{k} = "{v}"' for k, v in req.items() if k in ['phenotype', 'keyword']]

    columns = """in.rs_number as rs_number,
                 @class.asString() as relation_type,
                 phenotype,
                 clinical_significance,
                 out.namespace as namespace,
                 out.name as name"""

    from_where_sql = f""" FROM {relation_type}"""
    if wheres:
        from_where_sql += " WHERE " + ' AND '.join(wheres)

    number_of_results = b.query_get_dict(f"Select count(*) {from_where_sql}")[0]['count']

    sql = f"SELECT {columns} {from_where_sql} LIMIT {p.page_size} SKIP {p.skip}"

    return {
        'page': p.page,
        'page_size': p.page_size,
        'number_of_results': number_of_results,
        'pages': ceil(number_of_results / p.page_size),
        'results': b.query_get_dict(sql)
    }



"""BioGRID API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS, NUM_RESULTS, PAGE_SIZE


class TestBiogrid:
    """Unit tests for the Biogrid API module."""

    def test_get_has_ppi_bg_by_symbol_taxid(self, client):
        """Test has_ppi collection."""
        # TODO implement once fixed: currently returns an empty list
        pass

    def test_get_sources(self, client):
        response = client.get(
            'api/v1/biogrid/sources',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "BIOGRID": 1,
          "FLYBASE": 4,
          "POMBASE": 2,
          "WORMBASE": 3
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_has_ppi_bg_by_uniprot(self, client):
        response = client.get(
            'api/v1/ebel/biogrid/modification/by_uniprot?uniprot=P10636',
            content_type='application/json'
        )
        output = format_response_data(response)
        example_result = {  # Used to quickly getting cols
            "biogrid_ids": [
              2623252
            ],
            "modification": "Proteolytic Processing",
            "object_label": "Caspase-7",
            "object_name": "CASP7",
            "object_namespace": "HGNC",
            "object_taxonomy_id": 9606,
            "object_uniprot": "P55210",
            "pmids": [
              12888622
            ],
            "subject_label": "Microtubule-associated protein tau",
            "subject_name": "MAPT",
            "subject_namespace": "HGNC",
            "subject_taxonomy_id": 9606,
            "subject_uniprot": "P10636"
        }
        assert len(output) > 4
        assert isinstance(output, list)
        hit = None

        for entry in output:
            if entry["object_label"] == "Caspase-7":
                hit = entry
                break

        assert bool(hit)
        assert isinstance(hit, dict)
        for key, val in example_result.items():
            assert hit[key] == val

    def test_get_biogrid_by_biogrid_id(self, client):
        response = client.get(
            'api/v1/biogrid/by_biogrid_id/2658248',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "biogrid_a": {
            "entrez": 59272,
            "symbol": "ACE2",
            "systematic_name": "UNQ868/PRO1885",
            "taxonomy": {
              "organism_name": "Homo sapiens",
              "taxonomy_id": 9606
            },
            "trembl": None,
            "uniprot": "Q9BYF1"
          },
          "biogrid_b": {
            "entrez": 43740568,
            "symbol": "S",
            "systematic_name": "GU280_gp02",
            "taxonomy": {
              "organism_name": "Severe acute respiratory syndrome coronavirus 2",
              "taxonomy_id": 2697049
            },
            "trembl": None,
            "uniprot": "P0DTC2"
          },
          "biogrid_id": 2658248,
          "experimental_system": {
            "experimental_system": "Reconstituted Complex",
            "experimental_system_type": "physical",
            "frequency": 58466
          },
          "modification": None,
          "publication": {
            "author_name": "Sun C ",
            "publication_year": 2020,
            "source": "DOI",
            "source_identifier": "10.1101/2020.02.16.951723"
          },
          "qualification": "The Receptor Binding Domain (RBD) of the SARS-CoV-2 Spike protein interacts with human ACE2 in vitro.",
          "score": None,
          "source": "BIOGRID",
          "throughput": {
            "frequency": 366952,
            "throughput": "Low Throughput"
          }
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_biogrid_by_biogrid_id_using_post(self, client):
        response = client.post(
            'api/v1/biogrid/by_biogrid_id',
            json={"biogrid_id": 2658248},
        )
        output = format_response_data(response)

        expected_results = {
          "biogrid_a": {
            "entrez": 59272,
            "symbol": "ACE2",
            "systematic_name": "UNQ868/PRO1885",
            "taxonomy": {
              "organism_name": "Homo sapiens",
              "taxonomy_id": 9606
            },
            "trembl": None,
            "uniprot": "Q9BYF1"
          },
          "biogrid_b": {
            "entrez": 43740568,
            "symbol": "S",
            "systematic_name": "GU280_gp02",
            "taxonomy": {
              "organism_name": "Severe acute respiratory syndrome coronavirus 2",
              "taxonomy_id": 2697049
            },
            "trembl": None,
            "uniprot": "P0DTC2"
          },
          "biogrid_id": 2658248,
          "experimental_system": {
            "experimental_system": "Reconstituted Complex",
            "experimental_system_type": "physical",
            "frequency": 58466
          },
          "modification": None,
          "publication": {
            "author_name": "Sun C ",
            "publication_year": 2020,
            "source": "DOI",
            "source_identifier": "10.1101/2020.02.16.951723"
          },
          "qualification": "The Receptor Binding Domain (RBD) of the SARS-CoV-2 Spike protein interacts with human ACE2 in vitro.",
          "score": None,
          "source": "BIOGRID",
          "throughput": {
            "frequency": 366952,
            "throughput": "Low Throughput"
          }
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_experimental_systems(self, client):
        response = client.get(
            'api/v1/biogrid/experimental_systems',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "Affinity Capture-Luminescence": 24,
          "Affinity Capture-MS": 2,
          "Affinity Capture-RNA": 9,
          "Affinity Capture-Western": 5,
          "Biochemical Activity": 12,
          "Co-crystal Structure": 21,
          "Co-fractionation": 6,
          "Co-localization": 20,
          "Co-purification": 19,
          "Dosage Growth Defect": 26,
          "Dosage Lethality": 25,
          "Dosage Rescue": 18,
          "FRET": 22,
          "Far Western": 27,
          "Negative Genetic": 1,
          "PCA": 11,
          "Phenotypic Enhancement": 15,
          "Phenotypic Suppression": 14,
          "Positive Genetic": 4,
          "Protein-RNA": 17,
          "Protein-peptide": 23,
          "Proximity Label-MS": 7,
          "Reconstituted Complex": 8,
          "Synthetic Growth Defect": 10,
          "Synthetic Haploinsufficiency": 28,
          "Synthetic Lethality": 13,
          "Synthetic Rescue": 16,
          "Two-hybrid": 3
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_taxonomies(self, client):
        response = client.get(
            'api/v1/biogrid/taxonomies',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "Anopheles gambiae (PEST)": 180454,
          "Apis mellifera": 7460,
          "Arabidopsis thaliana (Columbia)": 3702,
          "Bacillus subtilis (168)": 224308,
          "Bos taurus": 9913,
          "Caenorhabditis elegans": 6239,
          "Candida albicans (SC5314)": 237561,
          "Canis familiaris": 9615,
          "Cavia porcellus": 10141,
          "Chlamydomonas reinhardtii": 3055,
          "Chlorocebus sabaeus": 60711,
          "Cricetulus griseus": 10029,
          "Danio rerio": 7955,
          "Dictyostelium discoideum (AX4)": 352472,
          "Drosophila melanogaster": 7227,
          "Emericella nidulans (FGSC A4)": 227321,
          "Equus caballus": 9796,
          "Escherichia coli (K12)": 83333,
          "Escherichia coli (K12/MC4100/BW2952)": 595496,
          "Escherichia coli (K12/MG1655)": 511145,
          "Escherichia coli (K12/W3110)": 316407,
          "Felis Catus": 9685,
          "Gallus gallus": 9031,
          "Glycine max": 3847,
          "Hepatitus C Virus": 11103,
          "Homo sapiens": 9606,
          "Human Herpesvirus 1": 10298,
          "Human Herpesvirus 2": 10310,
          "Human Herpesvirus 3": 10335,
          "Human Herpesvirus 4": 10376,
          "Human Herpesvirus 5": 10359,
          "Human Herpesvirus 6A": 32603,
          "Human Herpesvirus 6B": 32604,
          "Human Herpesvirus 7": 10372,
          "Human Herpesvirus 8": 37296,
          "Human Immunodeficiency Virus 1": 11676,
          "Human Immunodeficiency Virus 2": 11709,
          "Human papillomavirus (10)": 333759,
          "Human papillomavirus (16)": 333760,
          "Human papillomavirus (32)": 333763,
          "Human papillomavirus (5)": 333923,
          "Human papillomavirus (6b)": 10600,
          "Human papillomavirus (7)": 10620,
          "Human papillomavirus (9)": 10621,
          "Leishmania major (Friedlin)": 347515,
          "Macaca mulatta": 9544,
          "Meleagris gallopavo": 9103,
          "Middle-East Respiratory Syndrome-related Coronavirus": 1335626,
          "Monodelphis domestica": 13616,
          "Mus musculus": 10090,
          "Mycobacterium tuberculosis (H37Rv)": 83332,
          "Neurospora crassa (OR74A)": 367110,
          "Nicotiana tomentosiformis": 4098,
          "Oryctolagus cuniculus": 9986,
          "Oryza sativa (Japonica)": 39947,
          "Ovis aries": 9940,
          "Pan troglodytes": 9598,
          "Pediculus humanus": 121224,
          "Plasmodium falciparum (3D7)": 36329,
          "Rattus norvegicus": 10116,
          "Ricinus communis": 3988,
          "Saccharomyces cerevisiae (S288c)": 559292,
          "Schizosaccharomyces pombe (972h)": 284812,
          "Selaginella moellendorffii": 88036,
          "Severe acute respiratory syndrome coronavirus 2": 2697049,
          "Severe acute respiratory syndrome-related coronavirus": 694009,
          "Simian Immunodeficiency Virus": 11723,
          "Simian Virus 40": 10633,
          "Solanum lycopersicum": 4081,
          "Solanum tuberosum": 4113,
          "Sorghum bicolor": 4558,
          "Streptococcus pneumoniae (ATCCBAA255)": 171101,
          "Strongylocentrotus purpuratus": 7668,
          "Sus scrofa": 9823,
          "Tobacco Mosaic Virus": 12242,
          "Ustilago maydis (521)": 237631,
          "Vaccinia Virus": 10245,
          "Vitis vinifera": 29760,
          "Xenopus laevis": 8355,
          "Zea mays": 4577
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_modifications(self, client):
        response = client.get(
            'api/v1/biogrid/modifications',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "Acetylation": 4,
          "Deacetylation": 10,
          "Demethylation": 14,
          "Deneddylation": 15,
          "Dephosphorylation": 9,
          "Desumoylation": 13,
          "Deubiquitination": 5,
          "FAT10ylation": 19,
          "Glycosylation": 17,
          "Methylation": 7,
          "Nedd(Rub1)ylation": 11,
          "Neddylation": 18,
          "No Modification": 3,
          "Phosphorylation": 1,
          "Prenylation": 16,
          "Proteolytic Processing": 6,
          "Ribosylation": 12,
          "Sumoylation": 8,
          "Ubiquitination": 2,
          "de-ISGylation": 20
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_biogrid_by_pmid(self, client):
        response = client.get(
            'api/v1/biogrid/by_pmid/21685874',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
            "biogrid_a": {
              "entrez": 853167,
              "symbol": "GCN5",
              "systematic_name": "YGR252W",
              "taxonomy": {
                "organism_name": "Saccharomyces cerevisiae (S288c)",
                "taxonomy_id": 559292
              },
              "trembl": None,
              "uniprot": "Q03330"
            },
            "biogrid_b": {
              "entrez": 852295,
              "symbol": "HHT1",
              "systematic_name": "YBR010W",
              "taxonomy": {
                "organism_name": "Saccharomyces cerevisiae (S288c)",
                "taxonomy_id": 559292
              },
              "trembl": None,
              "uniprot": "P61830"
            },
            "biogrid_id": 543927,
            "experimental_system": {
              "experimental_system": "Biochemical Activity",
              "experimental_system_type": "physical",
              "frequency": 24247
            },
            "modification": {
              "frequency": 846,
              "modification": "Acetylation"
            },
            "publication": {
              "author_name": "Bian C ",
              "publication_year": 2011,
              "source": "PUBMED",
              "source_identifier": "21685874"
            },
            "qualification": "the purified SAGA complex can acetylate histone 3 in vitro",
            "score": None,
            "source": "BIOGRID",
            "throughput": {
              "frequency": 1611845,
              "throughput": "High Throughput"
            }
          }
        assert isinstance(output, list)
        hit = output[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_biogrid(self, client):
        response = client.get(
            'api/v1/biogrid?id_type_a=symbol&interactor_a=MAPT&taxonomy_id_a=9606&id_type_b=symbol&taxonomy_id_b=9606&modification=Acetylation&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "biogrid_a": {
            "entrez": 2033,
            "symbol": "EP300",
            "systematic_name": "RP1-85F18.1",
            "taxonomy": {
              "organism_name": "Homo sapiens",
              "taxonomy_id": 9606
            },
            "trembl": "Q7Z6C1",
            "uniprot": "Q09472"
          },
          "biogrid_b": {
            "entrez": 4137,
            "symbol": "MAPT",
            "systematic_name": None,
            "taxonomy": {
              "organism_name": "Homo sapiens",
              "taxonomy_id": 9606
            },
            "trembl": "B3KTM0",
            "uniprot": "P10636"
          },
          "biogrid_id": 558818,
          "experimental_system": {
            "experimental_system": "Biochemical Activity",
            "experimental_system_type": "physical",
            "frequency": 24247
          },
          "modification": {
            "frequency": 846,
            "modification": "Acetylation"
          },
          "publication": {
            "author_name": "Min SW ",
            "publication_year": 2010,
            "source": "PUBMED",
            "source_identifier": "20869593"
          },
          "qualification": None,
          "score": None,
          "source": "BIOGRID",
          "throughput": {
            "frequency": 366952,
            "throughput": "Low Throughput"
          }
        }
        assert output[NUM_RESULTS] == 4
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_biogrid_using_post(self, client):
        response = client.post(
            'api/v1/biogrid',
            json={
              "experimental_system_id": 2,
              "page": 2,
              "page_size": 3
            },
            follow_redirects=True,
        )
        output = format_response_data(response)

        expected_results = {
          "biogrid_a": {
            "entrez": 1489671,
            "symbol": "E",
            "systematic_name": "SARS-CoV-4",
            "taxonomy": {
              "organism_name": "Severe acute respiratory syndrome-related coronavirus",
              "taxonomy_id": 694009
            },
            "trembl": None,
            "uniprot": "P59637"
          },
          "biogrid_b": {
            "entrez": 83871,
            "symbol": "RAB34",
            "systematic_name": None,
            "taxonomy": {
              "organism_name": "Homo sapiens",
              "taxonomy_id": 9606
            },
            "trembl": "B4DNC0",
            "uniprot": "Q9BZG1"
          },
          "biogrid_id": 2754699,
          "experimental_system": {
            "experimental_system": "Affinity Capture-MS",
            "experimental_system_type": "physical",
            "frequency": 459921
          },
          "modification": None,
          "publication": {
            "author_name": "Stukalov A ",
            "publication_year": 2020,
            "source": "DOI",
            "source_identifier": "10.1101/2020.06.17.156455"
          },
          "qualification": "Affinity capture-MS was carried out using HA-tagged viral proteins as baits in human A549 lung carcinoma cells. Significant interactions were identified as those where the prey protein was at least 4 times enriched against the background (median log2 score > 2) with a p-value <= 1E-3.",
          "score": 8.15622,
          "source": "BIOGRID",
          "throughput": {
            "frequency": 1611845,
            "throughput": "High Throughput"
          }
        }
        assert output[NUM_RESULTS] > 459000
        assert output[PAGE_SIZE] == 3
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_interactor_by_symbol_starts_with(self, client):
        response = client.get(
            'api/v1/biogrid/interactor/by_symbol_starts_with?symbol=trem&taxonomy_id=9606&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "entrez": 54210,
          "symbol": "TREM1",
          "systematic_name": None,
          "taxonomy": {
            "organism_name": "Homo sapiens",
            "taxonomy_id": 9606
          },
          "trembl": "Q38L15",
          "uniprot": "Q9NP99"
        }
        assert output[NUM_RESULTS] == 4
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""Collection of methods for handling information caching."""
import re
import logging
import getpass
import pymysql

from lark.lexer import Token
from collections import defaultdict, Counter
from typing import Generator, List, Dict, DefaultDict, Set

from ebel import defaults
from ebel.manager import models
from ebel.tools import BelRdb
from ebel.config import write_to_config
from ebel.warnings import AlsoUsedInOtherNamespace, _Warning
from ebel.constants import URL, PATTERN, LIST, ALLOWED_TYPES
from ebel.constants import GRAMMAR_START_NS, GRAMMAR_START_ANNO
from ebel.errors import NotInNamespaceUrl, NotInAnnotationUrl, WithoutDefinedNamespace, \
    WithoutDefinedAnnotation, NotInNamespaceList, NotInAnnotationList, \
    NotInNamespacePattern, NotInAnnotationPattern, NotDownloadedFromUrl, _Error

# TODO: Because of change of `BelScript.ALLOWED_TYPES` FILE have to be handled on different way

logger = logging.getLogger(__name__)


def set_mysql_interactive() -> tuple:
    """Interactive mode to setup MySQL database."""
    print("Interactive mode\n \
          ================\n \
          1st setup of db and user with root:\n")
    host = input("host[localhost]:") or 'localhost'
    user = input("ebel user[ebel]:") or 'ebel'
    password = getpass.getpass(prompt="ebel password[ebel]:") or 'ebel'
    db = input("database name[ebel]") or 'ebel'
    print("If you want to setup the database automatically,\n \
          then type in the root password, otherwise nothing")
    root_pwd = getpass.getpass(prompt='root password (only for 1st setup):')

    if root_pwd:
        root_host = getpass.getpass(prompt='IP or name mysql server [localhost]:') or 'localhost'
        conn = pymysql.connect(host=root_host, user="root", password=root_pwd)
        c = conn.cursor()
        db_exists = c.execute("show databases like '{}'".format(db))

        if not db_exists:
            c.execute("CREATE DATABASE {} CHARACTER SET utf8".format(db))

        else:
            logger.warning(f"Database '{db}' already exists!")

        user_exists = c.execute("Select 1 from mysql.user where User='{}'".format(user))

        if not user_exists:
            sql = "CREATE USER '{}'@'{}' IDENTIFIED BY '{}'".format(
                user,
                host,
                password,
            )
            c.execute(sql)
        else:
            logger.warning(f"Database '{db}' already exists!")

        privileges_exists = c.execute("Select 1 from mysql.db where User='{}' and Db='{}'".format(user, db))
        if not privileges_exists:
            c.execute("GRANT ALL PRIVILEGES ON {}.* TO '{}'@'%'  IDENTIFIED BY '{}'".format(db, user, password))
        else:
            logger.warning(f"User already has privileges for database '{db}'")

        c.execute("FLUSH PRIVILEGES")

    return host, user, password, db


def set_mysql_connection(host: str = 'localhost',
                         user: str = 'ebel_user',
                         password: str = 'ebel_passwd',
                         db: str = 'ebel',
                         charset: str = 'utf8mb4'):
    """Set the connection using MySQL Parameters.

    Parameters
    ----------
    host : str
        MySQL database host.
    user : str
        MySQL database user.
    password : str
        MySQL database password.
    db : str
        MySQL database name.
    charset : str
        MySQL database charset.

    Returns
    -------
    str
        SQLAlchemy MySQL connection string.

    """
    connection_string = 'mysql+pymysql://{user}:{passwd}@{host}/{db}?charset={charset}'.format(
        host=host,
        user=user,
        passwd=password,
        db=db,
        charset=charset)
    set_connection(connection_string)

    return connection_string


def set_always_create_new_db(always_create_new_db: bool = True) -> None:
    """Set in configuration option `always_create_new_db`.

    Parameters
    ----------
    always_create_new_db : bool
        Option `always_create_new_db` in section `database` in config file.

    """
    write_to_config('database', 'always_create_new', str(always_create_new_db))


def set_connection(connection: str = defaults.CONN_STR_DEFAULT) -> None:
    """Set the connection string for SQLAlchemy.

    Parameters
    ----------
    connection: str
        SQLAlchemy connection string.

    """
    write_to_config('database', 'sqlalchemy_connection_string', connection)


class _BelScript:
    """Cache the content of the BEL script and methods to find errors and warnings."""

    def __init__(self, force_new_db):
        """Init."""
        # setup database
        engine = BelRdb().engine

        self.force_new_db = force_new_db
        models.reset_tables(engine, self.force_new_db)

        self._namespaces = Namespaces()  # entries Namespace objects
        self._annotations = Annotations()  # entries Annotation objects

        self.__namespace_in_db_updated = False
        self.__annotations_in_db_updated = False

        self._namespace_entries = NamespaceEntries()
        self._annotation_entries = AnnotationEntries()

        self.notDownloadedFromUrls = []

        self.namespace_manager = models.NamespaceManager(
            model=models.Namespace,
            entries_model=models.NamespaceEntry,
            grammar_start=GRAMMAR_START_NS
        )

        self.annotation_manager = models.AnnotationManager(
            model=models.Annotation,
            entries_model=models.AnnotationEntry,
            grammar_start=GRAMMAR_START_ANNO
        )

    def set_namespace_definition(self, as_type, keyword, value):
        """Set an annotation definition with type, keyword and value value could be 'file', 'url' or 'list'.

        :param str as_type: 'file', 'url' or 'list'
        :param str keyword: namespace keyword
        :param str value: URL, file path or list
        """
        if as_type in ALLOWED_TYPES:
            self._namespaces.add(as_type, keyword, value)
            return True
        else:
            logger.error("{} is not a allowed type of {}".format(as_type, ALLOWED_TYPES))
            return False

    def set_annotation_definition(self, as_type, keyword, value):
        """Set an annotation definition with type, keyword and value could be 'file', 'url' or 'list'.

        :param str as_type: 'file', 'url' or 'list'
        :param str keyword: namespace keyword
        :param str value: URL, file path or list
        """
        if as_type in ALLOWED_TYPES:
            self._annotations.add(as_type, keyword, value)
            return True
        else:
            logger.error("{} is not an allowed type of {}".format(as_type, ALLOWED_TYPES))
            return False

    def set_annotation_entry(self, annotation: str, entry: str, token: Token):
        """Set annotation, entry and lark.lexer.Token token.

        :param str annotation: annotation
        :param str entry: entry
        :param lark.lexer.Token token:
        """
        self._annotation_entries.set_annotation_entry(keyword=annotation, entry=entry, token=token)

    def set_namespace_entry(self, namespace: str, entry: str, token: Token):
        """Set namespace, entry and lark.lexer.Token token.

        :param str namespace:
        :param str entry:
        :param lark.lexer.Token token:
        """
        if not isinstance(token, Token):
            raise Exception("expecting Token in cache.set_namespace_entry")

        self._namespace_entries.set_namespace_entry(keyword=namespace, entry=entry, token=token)

    @property
    def errors(self) -> List[_Error]:
        """Execute all methods to find errors and warnings."""
        self.update_database()

        # all errors are children from errors._Error instances
        return (self.notDownloadedFromUrls
                + self.entries_without_namespace
                + self.entries_without_annotation
                + self.entries_not_in_namespace_url
                + self.entries_not_in_annotation_url
                + self.entries_not_in_namespace_list
                + self.entries_not_in_annotation_list
                + self.entries_not_in_namespace_pattern
                + self.entries_not_in_annotation_pattern
                )

    @property
    def warnings(self) -> List[_Warning]:
        """Execute all methods to find warnings."""
        if not (self.__namespace_in_db_updated and self.__annotations_in_db_updated):
            self.update_database()

        # all warnings are children from warnings._Warning instances
        return self.entries_also_in_other_namespace

    @property
    def entries_also_in_other_namespace(self) -> List[AlsoUsedInOtherNamespace]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        # extract all entries used in BEL statements and create a dict of lower entries with all keywords
        entry_keyword_dict = defaultdict(set)
        for keyword1, entries in self._namespace_entries.entries.items():
            for entry in entries:
                entry_keyword_dict[entry.lower()] |= {keyword1}
        # identify all ambiguous entries (in more than 1 namespace)
        ambiguous_entries = {entry: keywords for entry, keywords in entry_keyword_dict.items() if len(keywords) > 1}

        # ToDo: iterate all lower entries an check for permutation
#        for lower_entry in entry_keyword_dict:
#            if lower_entry.count(",") == 1:
#                reverse_without_comma = " ".join([x.strip() for x in lower_entry.split(",")][::-1])
#                if reverse_without_comma in entry_keyword_dict:
#                    print(lower_entry,
        #                    "%s exists in %s" % (reverse_without_comma,
        #                    entry_keyword_dict[reverse_without_comma]))
#                    ret.append(AlsoUsedInOtherNamespace(keyword=keyword2,
#                                                        entry=entry,
#                                                        line_number=token.line,
#                                                        column=token.column,
#                                                        hint=hint))

        # iterate all tokens with namespace entries and check if they are also exists in ambiguous entries
        for keyword2, entries_tokens in self._namespace_entries.tokens.items():
            for entry, tokens in entries_tokens.items():
                if entry.lower() in ambiguous_entries:
                    ambiguous_tokens = self._namespace_entries.tokens[keyword2][entry]
                    for token in ambiguous_tokens:
                        hint = "%s exists also in %s" % (entry, ambiguous_entries[entry.lower()] - {keyword2})
                        ret.append(AlsoUsedInOtherNamespace(keyword=keyword2,
                                                            entry=entry,
                                                            line_number=token.line,
                                                            column=token.column,
                                                            hint=hint))
        return ret

    @property
    def entries_not_in_namespace_pattern(self) -> List[NotInNamespacePattern]:
        """Return a list of entries not fitting a given namespace pattern."""
        ret = []

        ns_pattern_kwds = self.used_namespace_keywords & self._namespaces.keywords_by_type(PATTERN)

        for kwd in ns_pattern_kwds:
            regex = self._namespaces.keyword_dict[kwd].value
            pattern = re.compile("^" + regex + "$")
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if not pattern.search(entry):
                    ret.append(NotInNamespacePattern(keyword=kwd,
                                                     entry=entry,
                                                     line_number=line,
                                                     column=column))
        return ret

    @property
    def entries_not_in_annotation_pattern(self) -> List[NotInAnnotationPattern]:
        """Return a list of entries not fitting a given annotation pattern."""
        ret = []

        anno_pattern_kwds = self.used_annotation_keywords & self._annotations.keywords_by_type(PATTERN)

        for kwd in anno_pattern_kwds:
            regex = self._annotations.keyword_dict[kwd].value
            pattern = re.compile("^" + regex + "$")
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if not pattern.search(entry):
                    ret.append(NotInAnnotationPattern(keyword=kwd,
                                                      entry=entry,
                                                      line_number=line,
                                                      column=column))
        return ret

    @property
    def entries_not_in_annotation_list(self) -> List[NotInAnnotationList]:
        """Return a list of entries not in a given annotations."""
        ret = []

        anno_kwd_used_and_as_list = self.used_annotation_keywords & self._annotations.keywords_by_type(LIST)

        for kwd in anno_kwd_used_and_as_list:
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if entry not in self._annotations.keyword_dict[kwd].value:
                    ret.append(NotInAnnotationList(keyword=kwd,
                                                   entry=entry,
                                                   line_number=line,
                                                   column=column))
        return ret

    @property
    def entries_not_in_namespace_list(self) -> List[NotInNamespaceList]:
        """Return a list of entries not in a given namespace."""
        ret = []

        ns_kwd_used_and_as_list = self.used_namespace_keywords & self._namespaces.keywords_by_type(LIST)

        for kwd in ns_kwd_used_and_as_list:
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(kwd)
            for entry, line, column in elcs:
                if entry not in self._namespaces.keyword_dict[kwd].value:
                    ret.append(NotInNamespaceList(keyword=kwd,
                                                  entry=entry,
                                                  line_number=line,
                                                  column=column))
        return ret

    @property
    def entries_without_namespace(self) -> List[WithoutDefinedNamespace]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        for namespace_keyword in self.namespaces_without_definition:
            elcs = self._namespace_entries.get_entry_line_column_list_by_keyword(namespace_keyword)
            for entry, line, column in elcs:
                ret.append(WithoutDefinedNamespace(keyword=namespace_keyword,
                                                   entry=entry,
                                                   line_number=line,
                                                   column=column))
        return ret

    @property
    def entries_without_annotation(self) -> List[WithoutDefinedAnnotation]:
        """Return WithoutDefinedNamespace list."""
        ret = []
        for annotation_keyword in self.annotations_without_definition:
            elcs = self._annotation_entries.get_entry_line_column_list_by_keyword(annotation_keyword)
            for entry, line, column in elcs:
                ret.append(WithoutDefinedAnnotation(keyword=annotation_keyword,
                                                    entry=entry,
                                                    line_number=line,
                                                    column=column))
        return ret

    def update_database(self) -> None:
        """Update namespace and annotation entries in database if not exists by url and keyword."""
        if not (self.__namespace_in_db_updated and self.__annotations_in_db_updated):
            self.__namespace_in_db_updated = self.update_namespaces_in_db()
            self.__annotations_in_db_updated = self.update_annotations_in_db()

    def set_entry_not_in_namespace_list_errors(self):
        pass

    def set_entry_not_in_annotation_list_errors(self):
        pass

    @property
    def entries_not_in_namespace_url(self) -> List[NotInNamespaceUrl]:
        """Return a list of entries not exists in namespaces referenced as URL.

        Returns
        -------
        List[NotInNamespaceUrl]
            Description of returned object.
        """
        entries_not_in_namespace = []

        for keyword in self.used_namespace_keywords:

            namespace = self._namespaces.keyword_dict[keyword]

            if namespace.as_type == URL:

                url = namespace.value
                elc_list = self._namespace_entries.get_entry_line_column_list_by_keyword(keyword)

                names_not_exists = self.namespace_manager.get_entries_not_exists(
                    keyword=keyword,
                    url=url,
                    entry_line_column_list=elc_list,
                )

                for entry, line, column, hint in names_not_exists:
                    error = NotInNamespaceUrl(
                        keyword=keyword,
                        url_or_path=url,
                        entry=entry,
                        line_number=line,
                        column=column,
                        hint=hint
                    )
                    entries_not_in_namespace.append(error)

        return entries_not_in_namespace

    @property
    def entries_not_in_annotation_url(self) -> List[_Error]:
        """Return a list of entries not in the annotation URL."""
        entries_not_in_annotation = []

        for keyword in self.used_annotation_keywords:

            annotation = self._annotations.keyword_dict[keyword]

            if annotation.as_type == URL:

                url = annotation.value
                elc_list = self._annotation_entries.get_entry_line_column_list_by_keyword(keyword)

                names_not_exists = self.annotation_manager.get_entries_not_exists(
                    keyword=keyword,
                    url=url,
                    entry_line_column_list=elc_list
                )

                for entry, line, column, hint in names_not_exists:
                    error = NotInAnnotationUrl(
                        keyword=keyword,
                        url_or_path=url,
                        entry=entry,
                        line_number=line,
                        column=column,
                        hint=hint
                    )
                    entries_not_in_annotation.append(error)
        return entries_not_in_annotation

    def update_annotations_in_db(self) -> bool:
        """Update annotation in database if URL and keyword not exists."""
        import_success = True
        for anno in self._annotations.to_update:

            if anno.keyword in self.used_annotation_keywords:

                if not self.annotation_manager.keyword_url_exists(keyword=anno.keyword, url=anno.value):

                    if anno.as_type == URL:

                        logger.info(f"Update db with annotation {anno.keyword}: download from {anno.value}")

                        successful, error = self.annotation_manager.save_from_url_or_path(
                            keyword=anno.keyword,
                            url_or_path=anno.value,
                            doc_type=anno.as_type,
                        )

                        if not successful:
                            import_success = False
                            error_args = error.args[0].split("\n")
                            string_error = error_args[2] if len(error_args) > 1 else error_args[0]
                            logger.error(f"Annotation {anno.keyword} failed to be added from {anno.value}",
                                         exc_info=False)

                            if "column" in dir(error):  # Indicates it's a Lark error
                                download_error = NotDownloadedFromUrl(
                                    keyword=anno.keyword,
                                    url_or_path=anno.value,
                                    column=error.column,
                                    line=error.line,
                                    hint=f'{error.allowed} error in "{string_error}"'
                                )

                            else:  # It's an HTTPError of some kind
                                download_error = NotDownloadedFromUrl(
                                    keyword=anno.keyword,
                                    url_or_path=anno.value,
                                    column=0,
                                    line=0,
                                    hint=f"{string_error}"
                                )
                            self.notDownloadedFromUrls.append(download_error)

        return import_success

    def update_namespaces_in_db(self) -> bool:
        """Update namespaces in database if URL and keyword does not exist."""
        import_success = True
        for ns in self._namespaces.to_update:

            if ns.keyword in self.used_namespace_keywords:

                if not self.namespace_manager.keyword_url_exists(keyword=ns.keyword, url=ns.value):

                    if ns.as_type == URL:

                        logger.info(f"Update db with namespace {ns.keyword}: download from {ns.value}")

                        successful, error = self.namespace_manager.save_from_url_or_path(
                            keyword=ns.keyword,
                            url_or_path=ns.value,
                            doc_type=ns.as_type,
                        )

                        if not successful:
                            import_success = False
                            error_args = error.args[0].split("\n")
                            string_error = error_args[2] if len(error_args) > 1 else error_args[0]
                            logger.error(f"Namespace {ns.keyword} failed to be added from {ns.value}",
                                         exc_info=False)

                            if "column" in dir(error):  # Indicates it's a Lark error
                                download_error = NotDownloadedFromUrl(
                                    keyword=ns.keyword,
                                    url_or_path=ns.value,
                                    column=error.column,
                                    line=error.line,
                                    hint=f'{error.allowed} error in "{string_error}"'
                                )

                            else:  # It's an HTTPError of some kind
                                download_error = NotDownloadedFromUrl(
                                    keyword=ns.keyword,
                                    url_or_path=ns.value,
                                    column=0,
                                    line=0,
                                    hint=f"{string_error}"
                                )

                            self.notDownloadedFromUrls.append(download_error)

        return import_success

    @property
    def namespaces_with_multiple_definitions(self):
        """Return all Namespace objects with several definitions.

        This is returned as a dictionary (key:keyword, value: list of Namespace objects).
        """
        ret = defaultdict(list)
        multiple_keyword = [k for k, v in Counter(self._namespaces.keywords).items() if v > 1]
        for ns in self._namespaces:
            if ns.keyword in multiple_keyword:
                ret[ns.keyword].append(ns)
        return dict(ret)

    @property
    def annotations_with_multiple_definitions(self):
        """Return all Annotation objects with several definitions.

        This is returned as a dictionary (key:keyword, value: list of Annotation objects).
        """
        ret = defaultdict(list)
        multiple_keyword = [k for k, v in Counter(self._annotations.keywords).items() if v > 1]
        for anno in self._annotations:
            if anno.keyword in multiple_keyword:
                ret[anno.keyword].append(anno)
        return dict(ret)

    @property
    def namespaces_without_definition(self):
        """Return set of namespace keywords used in statements but not defined with a reference.

        :return set: set of str
        """
        return set(self._namespace_entries.keywords) - set(self._namespaces.keywords)

    @property
    def annotations_without_definition(self):
        """Return a set of annotation keywords not defined with a reference.

        :return set: set of str
        """
        return set(self._annotation_entries.keywords) - set(self._annotations.keywords)

    @property
    def used_namespace_keywords(self) -> Set[str]:
        """Return set of used namespace keywords (with reference and used in statements)."""
        return set(self._namespace_entries.keywords) & set(self._namespaces.keywords)

    @property
    def used_annotation_keywords(self) -> Set[str]:
        """Return set of used namespace keywords."""
        return set(self._annotation_entries.keywords) & set(self._annotations.keywords)

    @property
    def namespace_keywords_in_statements(self):
        """Return all unique namespace keywords used in statements."""
        return self._namespace_entries.keywords

    @property
    def annotation_keywords_in_statements(self):
        """Return all unique annotation keywords used in statements."""
        return self._namespace_entries.keywords

    def get_entries_by_namespace_keyword(self, keyword):
        """Get all entries by namespace keyword.

        :param keyword: namespace keyword
        :return set: all entries in the namespace
        """
        return self._namespace_entries.get_entries_by_keyword(keyword)

    def get_entries_by_annotation_keyword(self, keyword):
        """Get all entries by namespace keyword.

        :param keyword: namespace keyword
        :return set: all entries in the namespace
        """
        return self._annotation_entries.get_entries_by_keyword(keyword)


class Entries:
    """Abstract class representing namespaces and annotations."""

    tokens = defaultdict(dict)
    entries = defaultdict(set)

    def get_entry_line_column_list_by_keyword(self, keyword: str) -> Generator[str, int, int]:
        """Get generator of tuple(entry, line, column) by keyword.

        Parameters
        ----------
        keyword: str
            Description of parameter `keyword: str`.

        Returns
        -------
        Generator
            Generator of tuple(entry: str, line: int, column: int).
        """
        for entry, tokens in self.tokens[keyword].items():
            for token in tokens:
                yield entry, token.line, token.column

    @property
    def keywords(self):
        """Return a list of unique keywords used in SETs."""
        return self.entries.keys()

    def get_entries_by_keyword(self, keyword: str) -> Set:
        """Get entries by keyword.

        :param str keyword: keyword to retrieve from dict
        """
        return self.entries.get(keyword, set())

    def get_tokens_by_keyword(self, keyword: str) -> Dict:
        """Get tokens by keyword.

        :param str keyword: keyword to retrieve from dict
        """
        return self.entries.get(keyword, set())

    def __str__(self):
        """String representation of object."""
        return str(dict(self.tokens))


class NamespaceEntries(Entries):
    """Namespace subclass of Entries."""

    def __init__(self):
        """Init."""
        self.entries = defaultdict(set)
        self.tokens = defaultdict(dict)

    def set_namespace_entry(self, keyword, entry, token):
        """Set namespace,  entry and lark.lexer.Token.

        :param str keyword: namespace
        :param str entry: entry
        :param lark.lexer.Token token: Token object from lark library
        """
        if isinstance(token, Token):
            self.entries[keyword] |= {entry}
            if keyword in self.tokens and entry in self.tokens[keyword]:
                self.tokens[keyword][entry].append(token)
            else:
                self.tokens[keyword][entry] = [token]
        else:
            raise "Argument token is type {} not {}".format(type(token), 'lark.lexer.Token')


class AnnotationEntries(Entries):
    """Annotation subclass of Entries."""

    def __init__(self):
        """Init."""
        self.entries = defaultdict(set)
        self.tokens = defaultdict(dict)

    def set_annotation_entry(self, keyword: str, entry, token):
        """Set annotation,  entry and lark.lexer.Token.

        :param str keyword: annotation
        :param entry: entry
        :param lark.lexer.Token token: Token object from lark library
        """
        if isinstance(token, Token):
            self.entries[keyword] |= {entry}
            if keyword in self.tokens and entry in self.tokens[keyword]:
                self.tokens[keyword][entry].append(token)
            else:
                self.tokens[keyword][entry] = [token]
        else:
            raise "argument token is type {} not {}".format(type(token), 'lark.lexer.Token')


class NsAnsBase:
    """Parent class for class Namespace and Annotation."""

    def __init__(self, obj_class):
        """Init."""
        self.__objs = []
        self.class_ = obj_class

    def add(self, as_type: str, keyword: str, value: str):
        """Add obj to list of objs.

        :param str as_type: allowed keywords 'file', 'url' or 'list'
        :param str keyword: keyword used in object
        :param str value: value of object
        :return:
        """
        obj = self.class_(as_type, keyword, value)
        self.__objs.append(obj)

    @property
    def type_dict(self) -> DefaultDict:
        """Convert to list of dictionaries."""
        ret = defaultdict(list)
        [ret[obj.as_type].append(obj) for obj in self]
        return ret

    def by_type(self, as_type: str):
        """Return list of Namespace objects by 'list', 'url' or 'file'."""
        if as_type not in ALLOWED_TYPES:
            raise "{} not in allowed types {}".format(as_type, ALLOWED_TYPES)
        return [obj for obj in self if obj.as_type == as_type]

    def keywords_by_type(self, as_type: str) -> Set[str]:
        """Return a set of keywords by Namespace type 'list', 'url' or 'file'."""
        if as_type not in ALLOWED_TYPES:
            raise "{} not in allowed types {}".format(as_type, ALLOWED_TYPES)
        return set([obj.keyword for obj in self if obj.as_type == as_type])

    @property
    def keyword_dict(self) -> Dict:
        """Return a dictionary of key=keyword, value: Namespace or Annotation object."""
        ret = dict()
        for obj in self:
            ret[obj.keyword] = obj
        return ret

    @property
    def keywords(self) -> List[str]:
        """Return all keywords."""
        return [obj.keyword for obj in self.__objs]

    @property
    def to_update(self) -> List:
        """Return a list of all Namespace or Annotation (NS_or_Anno) objects with URL or file path.

        :return list: list of all Namespace or Annotation (NS_or_Anno) objects with URL or file path
        """
        return self.type_dict[URL]

    def __iter__(self):
        """Return a generator of objects (Namespace or Annotation)."""
        for obj in self.__objs:
            yield obj


class Namespaces(NsAnsBase):
    """Namespace child class."""

    def __init__(self):
        """init."""
        super(Namespaces, self).__init__(obj_class=Namespace)


class Annotations(NsAnsBase):
    """Annotation child class."""

    def __init__(self):
        """init."""
        super(Annotations, self).__init__(obj_class=Annotation)


class Namespace:
    """Namespace class to represent BEL statement namespaces."""

    def __init__(self, as_type, keyword, value):
        """Namespace init."""
        self.as_type = as_type
        self.keyword = keyword
        self.value = value

    def to_dict(self):
        """Convert class values to dictionary."""
        return {'as_type': self.as_type, 'keyword': self.keyword, 'value': self.value}

    def __unicode__(self):
        return "Namespace:" + str(self.to_dict())

    def __str__(self):
        return self.__unicode__()


class Annotation:
    """Annotation class to represent BEL statement annotations."""

    def __init__(self, as_type, keyword, value):
        """Annotation init."""
        self.as_type = as_type
        self.keyword = keyword
        self.value = value

    def to_dict(self):
        """Convert class values to dictionary."""
        return {'as_type': self.as_type, 'keyword': self.keyword, 'value': self.value}

    def __unicode__(self):
        return "Annotation" + str(self.to_dict())

    def __str__(self):
        return self.__unicode__()



"""HGNC API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS, NUM_RESULTS, PAGE_SIZE


class TestHgnc:
    def test_get_by_symbol(self, client):
        response = client.get(
            'api/v1/hgnc/by_symbol?symbol=CD33',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
            "agr": "HGNC:1659",
            "alias_names": [
                "sialic acid binding Ig-like lectin 3"
            ],
            "alias_symbols": [
                "SIGLEC3",
                "SIGLEC-3",
                "p67",
                "FLJ00391"
            ],
            "bioparadigms_slc": None,
            "ccdss": [
                "CCDS46157",
                "CCDS54299",
                "CCDS33084"
            ],
            "cd": "CD33",
            "cosmic": None,
            "date_approved_reserved": "1986-01-01",
            "date_modified": "2016-10-05",
            "date_name_changed": "2006-03-28",
            "date_symbol_changed": None,
            "enas": [
                "M23197"
            ],
            "ensembl_gene_id": "ENSG00000105383",
            "entrez_id": 945,
            "enzymes": [],
            "gene_group_ids": [
                471,
                590,
                745
            ],
            "gene_group_names": [
                "CD molecules",
                "V-set domain containing",
                "Sialic acid binding Ig like lectins"
            ],
            "hgnc_id": "HGNC:1659",
            "homeodb": None,
            "horde_id": None,
            "imgt": None,
            "intermediate_filament_db": None,
            "iuphar": "objectId:2601",
            "kznf_gene_catalog": None,
            "lncipedia": None,
            "lncrnadb": None,
            "location": "19q13.41",
            "location_sortable": "19q13.41",
            "locus_group": "protein-coding gene",
            "locus_type": "gene with protein product",
            "lsdbs": [],
            "mamit_trnadb": None,
            "merops": None,
            "mgds": [
                "MGI:99440"
            ],
            "mirbase": None,
            "name": "CD33 molecule",
            "omims": [
                159590
            ],
            "orphanet": None,
            "pre_symbols": [],
            "prev_names": [
                "CD33 antigen (gp67)"
            ],
            "pseudogene_org": None,
            "pubmeds": [
                3139766,
                9465907
            ],
            "refseqs": [
                "NM_001772"
            ],
            "rgds": [
                "RGD:1596020"
            ],
            "rna_centrals": [],
            "snornabase": None,
            "status": "Approved",
            "symbol": "CD33",
            "ucsc_id": "uc002pwa.3",
            "uniprots": [
                "P20138"
            ],
            "uuid": "982e665a-2cc7-4ef3-b9f6-4406e08ed6c8",
            "vega_id": "OTTHUMG00000182891",
            "version": 1702607059790331904
        }
        assert isinstance(output, dict)
        assert output == expected_results

    def test_get_uniprot_accession_by_hgnc_symbol(self, client):
        response = client.get(
            'api/v1/hgnc/uniprot_accession_by_hgnc_symbol?symbol=CD33',
            content_type='application/json'
        )
        output = format_response_data(response)
        assert isinstance(output, str)
        assert output == "P20138"

    def test_get_hgnc(self, client):
        response = client.get(
            'api/v1/hgnc?symbol=CD33&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "agr": "HGNC:1659",
          "alias_names": [
            "sialic acid binding Ig-like lectin 3"
          ],
          "alias_symbols": [
            "SIGLEC3",
            "SIGLEC-3",
            "p67",
            "FLJ00391"
          ],
          "bioparadigms_slc": None,
          "ccdss": [
            "CCDS46157",
            "CCDS54299",
            "CCDS33084"
          ],
          "cd": "CD33",
          "cosmic": None,
          "date_approved_reserved": "1986-01-01",
          "date_modified": "2016-10-05",
          "date_name_changed": "2006-03-28",
          "date_symbol_changed": None,
          "enas": [
            "M23197"
          ],
          "ensembl_gene_id": "ENSG00000105383",
          "entrez_id": 945,
          "enzymes": [],
          "gene_group_ids": [
            471,
            590,
            745
          ],
          "gene_group_names": [
            "CD molecules",
            "V-set domain containing",
            "Sialic acid binding Ig like lectins"
          ],
          "hgnc_id": "HGNC:1659",
          "homeodb": None,
          "horde_id": None,
          "imgt": None,
          "intermediate_filament_db": None,
          "iuphar": "objectId:2601",
          "kznf_gene_catalog": None,
          "lncipedia": None,
          "lncrnadb": None,
          "location": "19q13.41",
          "location_sortable": "19q13.41",
          "locus_group": "protein-coding gene",
          "locus_type": "gene with protein product",
          "lsdbs": [],
          "mamit_trnadb": None,
          "merops": None,
          "mgds": [
            "MGI:99440"
          ],
          "mirbase": None,
          "name": "CD33 molecule",
          "omims": [
            159590
          ],
          "orphanet": None,
          "pre_symbols": [],
          "prev_names": [
            "CD33 antigen (gp67)"
          ],
          "pseudogene_org": None,
          "pubmeds": [
            3139766,
            9465907
          ],
          "refseqs": [
            "NM_001772"
          ],
          "rgds": [
            "RGD:1596020"
          ],
          "rna_centrals": [],
          "snornabase": None,
          "status": "Approved",
          "symbol": "CD33",
          "ucsc_id": "uc002pwa.3",
          "uniprots": [
            "P20138"
          ],
          "uuid": "982e665a-2cc7-4ef3-b9f6-4406e08ed6c8",
          "vega_id": "OTTHUMG00000182891",
          "version": 1702607059790331904
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results



"""miRTarBase API methods."""

from flask import request

from ebel.manager.rdbms.models import mirtarbase
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_mirtarbase():
    """Get generic mirtarbase entry."""
    return _get_data(mirtarbase.Mirtarbase)


def get_ebel_relation():
    """Get miRTarBase related eBEL relations."""
    conf = {
        'mirbase_id': "out.name",
        'target_rna_symbol': "in.name",
        'target_namespace': "in.namespace",
        'support_type': "support_type",
        'pmid': "pmid",
        'experiments': "experiments"
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_mirgene_target"

    ra = request.args

    wheres = []

    experiment = ra.get('experiment')
    if experiment:
        wheres.append(f'"{experiment}" in experiments')

    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        wheres += [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]

    if wheres:
        sql += " WHERE " + ' AND '.join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""API methods."""
import connexion
import webbrowser

from flask_cors import CORS

application = connexion.FlaskApp(__name__)
application.add_api('openapi.yml')

CORS(application.app, expose_headers=["Content-Disposition"])


def run(host: str = '0.0.0.0', port: int = 5000, debug_mode: bool = True, open_browser: bool = False):
    """Run the API server.

    Parameters
    ----------
    host: str
        Server or host for the API service.
    port: int
        The defined port to run the API server.
    debug_mode: bool
        Whether bebug mode should be enabled.
    open_browser: bool
        If True, automatically opens browser to API UI.
    """
    url = f'http://{host}:{port}/ui'
    if open_browser:
        webbrowser.open(url)
    print(f'Starting web server {url}')
    application.run(host=host, port=port, debug=debug_mode)


if __name__ == '__main__':
    run()



"""OrientDB structure.

All vertex, edge and index definitions have to follow the following conversion:

prefix: Used in the hole package for domains like BEL, HGNC, UniProt, ...
typeOfClass or Index
"""
from copy import deepcopy
from enum import Enum
from typing import List, Dict, Optional, Tuple

from ebel.manager.orientdb.odb_defaults import OIndexType, ODataType, normalized_pmod


class OClassType(Enum):
    """Class for constants definitions."""

    NODE = 'VERTEX'
    EDGE = 'EDGE'
    GENERIC = 'GENERIC'


class OProperty(object):
    """Generic class definition for creating properties in the OrientDB database."""

    def __init__(self,
                 prop_name: str,
                 data_type: ODataType,
                 linked_class: Optional[str] = None,
                 linked_type: Optional[ODataType] = None,
                 mandatory: bool = False,
                 node_view_label=False,
                 node_view_sub_label=False):
        """Init method."""
        self.prop_name = prop_name
        self.data_type = data_type
        self.linked_type = linked_type
        self.linked_class = linked_class
        self.mandatory = mandatory
        self.node_view_label = node_view_label
        self.node_view_sub_label = node_view_sub_label


class OClass(object):
    """Generic class definition for creating classes in the OrientDB database."""

    def __init__(self,
                 name: str,
                 extends: Tuple[str, ...],
                 abstract: bool,
                 props: Tuple[OProperty, ...],
                 own_class: bool,
                 class_type: OClassType):
        """Init method."""
        self.class_type = class_type
        self.name = name
        self.extends = extends
        self.abstract = abstract
        self.props = list(props)
        self.own_class = own_class

    def add_properties(self, props: List[OProperty]):
        """Add properties to the class."""
        for prop in props:
            self.props.append(prop)
        return self

    def is_edge(self):
        """Check whether class is an edge."""
        return self.class_type == OClassType.EDGE

    def is_node(self):
        """Check whether class is a vertex/node."""
        return self.class_type == OClassType.NODE

    def is_vertex(self):
        """Check whether class is a vertex. Returns "is_node" method and created for convenience."""
        return self.is_node()

    def is_generic(self):
        """Check whether class is generic."""
        return self.class_type == OClassType.GENERIC


class Node(OClass):
    """Generic class definition for creating node classes in the OrientDB database."""

    def __init__(self,
                 name: str,
                 extends: Tuple[OClass, ...] = (),
                 abstract: bool = False,
                 props: Tuple[OProperty, ...] = (),
                 own_class: bool = True):
        """Init method for Node ODB class."""
        extends = tuple(x.name for x in extends)
        OClass.__init__(self, name, extends, abstract, props, own_class, class_type=OClassType.NODE)


class Edge(OClass):
    """Generic class definition for creating edge classes in the OrientDB database."""

    def __init__(self,
                 name: str,
                 extends: Tuple[OClass, ...] = (),
                 abstract: bool = False,
                 props: Tuple[OProperty, ...] = (),
                 own_class: bool = True,
                 in_out: Tuple[Optional[OClass], Optional[OClass]] = (None, None)):
        """Init method for Edge ODB class."""
        extends = tuple(x.name for x in extends)
        in_ = in_out[0].name if in_out[0] else None
        out = in_out[1].name if in_out[1] else None
        OClass.__init__(self, name, extends, abstract, props, own_class, class_type=OClassType.EDGE)
        self.in_out = (in_, out)


class Generic(OClass):
    """Generic class definition for creating generic classes in the OrientDB database."""

    def __init__(self,
                 name: str,
                 extends: Tuple[str, ...] = (),
                 abstract: bool = False,
                 props: Tuple[OProperty, ...] = (),
                 own_class: bool = True):
        """Init method for Generic ODB class."""
        OClass.__init__(self, name, extends, abstract, props, own_class, class_type=OClassType.GENERIC)


class OIndex(object):
    """Generic class definition for creating indices in the OrientDB database."""

    def __init__(self, odb_class: OClass, columns: Tuple[str, ...], index_type: OIndexType):
        """Init method."""
        self.class_name = odb_class.name
        self.columns = columns
        self.index_type = index_type


basic_node = Node(name='V')
basic_edge = Edge(name='E')

##############################################################################
# Definition of BEL vertices, edges and indices
##############################################################################
bel_document = Generic('bel_document', props=(
    OProperty('name', ODataType.STRING),
    OProperty('date_uploaded', ODataType.DATETIME),
    OProperty('description', ODataType.STRING),
    OProperty('version', ODataType.STRING),
    OProperty('authors', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('contact_info', ODataType.STRING),
    OProperty('copyright', ODataType.STRING),
    OProperty('licences', ODataType.STRING),
    OProperty('git', ODataType.EMBEDDEDMAP),
    OProperty('namespaces', ODataType.EMBEDDEDMAP),
    OProperty('annotations', ODataType.EMBEDDEDMAP),
    OProperty('file', ODataType.EMBEDDEDMAP),
    OProperty('git_info', ODataType.EMBEDDEDMAP),
    OProperty('keywords', ODataType.LINKSET, 'keyword'),
))

bel_generics: Tuple[Generic, ...] = (
    Generic('keyword', props=(
        OProperty('label', ODataType.STRING),
        OProperty('description', ODataType.STRING),
    )),
    bel_document,
)
namespace_name = Node('nn', (basic_node,), abstract=True, props=(
    OProperty('namespace', ODataType.STRING),
    OProperty('name', ODataType.STRING),
))
bel = Node('bel', (basic_node,), abstract=True, props=(
    OProperty('bel', ODataType.STRING, mandatory=True, node_view_sub_label=True),
    OProperty('label', ODataType.STRING),
    OProperty('involved_genes', ODataType.EMBEDDEDSET, linked_type=ODataType.STRING, node_view_label=True),
    OProperty('involved_other', ODataType.EMBEDDEDSET, linked_type=ODataType.STRING, node_view_label=True),
    OProperty('suggested_corrections', ODataType.EMBEDDEDMAP, linked_type=ODataType.STRING),
))
bio_object = Node('bio_object', (bel,), abstract=True, props=(
    OProperty('chebi', ODataType.INTEGER),
))
pure_object = Node('pure_object', (basic_node,), abstract=True, props=(
    OProperty('pure', ODataType.BOOLEAN),
    OProperty('species', ODataType.INTEGER),
))
genetic_flow = Node('genetic_flow', (bio_object, namespace_name, pure_object), abstract=True)
bio_concept = Node('bio_concept', (bel,), abstract=True)
location_object = Node('location_object', (basic_node,), abstract=True, props=(
    OProperty('location', ODataType.EMBEDDEDMAP),
))
bio_act = Node('bio_act', (bel,))
bio_list = Node('bio_list', (bel,), abstract=True)
ebel = Node('ebel', (basic_node,), abstract=True, props=(
    OProperty('bel', ODataType.STRING, node_view_sub_label=True),
    OProperty('name', ODataType.STRING, node_view_label=True),
    OProperty('namespace', ODataType.STRING),
    OProperty('label', ODataType.STRING),
))
protein = Node('protein', (genetic_flow, location_object), props=(
    OProperty('uniprot', ODataType.STRING),
))
gene = Node('gene', (genetic_flow, location_object))
rna = Node('rna', (genetic_flow, location_object))
abundance = Node('abundance', (bio_object, namespace_name, pure_object))
population = Node('population', (bio_object, namespace_name, pure_object))
from_location = Node('from_location', (ebel,))
to_location = Node('to_location', (ebel,))
biological_process = Node('biological_process', (bio_concept, namespace_name))
pathology = Node('pathology', (bio_concept, namespace_name))
complex_ = Node('complex', (bio_object, namespace_name, pure_object))
micro_rna = Node('micro_rna', (bio_object, namespace_name, location_object))
activity = Node('activity', (bio_act, namespace_name), props=(
    OProperty('default', ODataType.STRING),
))
reaction = Node('reaction', (bio_act,))
degradation = Node('degradation', (bio_act,))
cell_secretion = Node('cell_secretion', (bio_act,))
translocation = Node('translocation', (bio_act,))
cell_surface_expression = Node('cell_surface_expression', (bio_act,))
list_ = Node('list', (bio_list,))
composite = Node('composite', (bio_list,))
variant = Node('variant', (ebel,), props=(
    OProperty('hgvs', ODataType.STRING, node_view_label=True),
))
fragment = Node('fragment', (ebel,))
location = Node('location', (ebel,))
pmod = Node('pmod', (ebel,), props=(
    OProperty('position', ODataType.INTEGER, node_view_label=True),
    OProperty('type', ODataType.STRING, node_view_label=True),
    OProperty('amino_acid', ODataType.STRING, node_view_label=True),
))
gmod = Node('gmod', (ebel,))
reactants = Node('reactants', (ebel,))
products = Node('products', (ebel,))
fusion_protein = Node('fusion_protein', (bel,))
fusion_rna = Node('fusion_rna', (bel,))
fusion_gene = Node('fusion_gene', (bel,))

bel_nodes: Tuple[Node, ...] = (
    bel,
    namespace_name,
    pure_object,
    location_object,
    bio_concept,
    biological_process,
    pathology,
    bio_object,
    complex_,
    abundance,
    population,
    genetic_flow,
    gene,
    rna,
    protein,
    micro_rna,
    bio_act,
    activity,
    reaction,
    degradation,
    cell_secretion,
    translocation,
    cell_surface_expression,
    bio_list,
    list_,
    composite,
    ebel,
    variant,
    fragment,
    location,
    pmod,
    gmod,
    from_location,
    to_location,
    reactants,
    products,
    fusion_protein,
    fusion_rna,
    fusion_gene,
)

bel_relation = Edge('bel_relation', (basic_edge,), abstract=True, props=(
    OProperty("evidence", ODataType.STRING),
    OProperty("pmid", ODataType.INTEGER),
    OProperty("pmc", ODataType.STRING),
    OProperty("citation", ODataType.EMBEDDEDMAP),
    OProperty("annotation", ODataType.EMBEDDEDMAP),
    OProperty("document", ODataType.LINKSET, linked_class=bel_document.name)))
ebel_relation = Edge('ebel_relation', (basic_edge,), abstract=True)
causal = Edge('causal', (bel_relation,), abstract=True, in_out=(bel, bel))
correlative = Edge('correlative', (bel_relation,), abstract=True, in_out=(bel, bel))
genomic = Edge('genomic', (bel_relation,), abstract=True)
other = Edge('other', (bel_relation,), abstract=True)
deprecated = Edge('deprecated', (bel_relation,), abstract=True)
compiler = Edge('compiler', (bel_relation,), abstract=True)
has_modified = Edge('has_modified', (ebel_relation,), abstract=True)
has_variant_obj = Edge('has_variant_obj', (ebel_relation,), abstract=True)
has_located = Edge('has_located', (ebel_relation,), abstract=True)
has_ppi = Edge('has_ppi', (ebel_relation,), abstract=True, own_class=False, in_out=(protein, protein))

bel_edges: Tuple[Edge, ...] = (
    bel_relation,
    causal,
    Edge('increases', (causal,)),
    Edge('directly_increases', (causal,)),
    Edge('decreases', (causal,)),
    Edge('directly_decreases', (causal,)),
    Edge('rate_limiting_step_of', (causal,)),
    Edge('causes_no_change', (causal,)),
    Edge('regulates', (causal,)),

    correlative,
    Edge('negative_correlation', (correlative,)),
    Edge('positive_correlation', (correlative,)),
    Edge('association', (correlative,)),
    Edge('no_correlation', (correlative,)),

    genomic,
    Edge('orthologous', (genomic,), in_out=(bel, bel)),
    Edge('transcribed_to', (genomic,), in_out=(rna, gene)),
    Edge('translated_to', (genomic,), in_out=(protein, rna)),

    other,
    Edge('has_member', (other,), in_out=(bel, bel)),
    Edge('has_members', (other,), in_out=(bel, bel)),
    Edge('has_component', (other,), in_out=(bel, bel)),
    Edge('has_components', (other,), in_out=(bel, bel)),
    Edge('equivalent_to', (other,), in_out=(bel, bel)),
    Edge('is_a', (other,), in_out=(bel, bel)),
    Edge('sub_process_of', (other,), in_out=(bel, bel)),

    deprecated,
    Edge('analogous_to', (deprecated,), in_out=(bel, bel)),
    Edge('biomarker_for', (deprecated,), in_out=(bel, bel)),
    Edge('prognostic_biomarker_for', (deprecated,)),

    compiler,
    Edge('acts_in', (compiler,), in_out=(bel, bel)),
    Edge('has_product', (compiler,), in_out=(bel, bel)),
    Edge('has_variant', (compiler,), in_out=(bel, bel)),
    Edge('has_modification', (compiler,), in_out=(bel, bel)),
    Edge('reactant_in', (compiler,), in_out=(bel, bel)),
    Edge('translocates', (compiler,), in_out=(bel, bel)),
    Edge('includes', (compiler,), in_out=(bel, bel)),

    ebel_relation,
    # TODO: check if always basic_node is needed or if we can refine that
    Edge('has__protein', (ebel_relation,), in_out=(protein, basic_node)),
    Edge('has__rna', (ebel_relation,), in_out=(rna, basic_node)),
    Edge('has__gene', (ebel_relation,), in_out=(gene, basic_node)),
    Edge('has__abundance', (ebel_relation,), in_out=(abundance, basic_node)),
    Edge('has__population', (ebel_relation,), in_out=(population, basic_node)),
    Edge('has__location', (ebel_relation,), in_out=(ebel, basic_node)),
    Edge('has__from_location', (ebel_relation,), in_out=(from_location, basic_node)),
    Edge('has__to_location', (ebel_relation,), in_out=(to_location, basic_node)),
    Edge('has__fragment', (ebel_relation,), in_out=(fragment, basic_node)),
    Edge('has__pmod', (ebel_relation,), in_out=(pmod, basic_node)),
    Edge('has__gmod', (ebel_relation,), in_out=(gmod, basic_node)),
    Edge('has__complex', (ebel_relation,), in_out=(complex_, basic_node)),
    Edge('has__micro_rna', (ebel_relation,), in_out=(micro_rna, basic_node)),
    Edge('has__variant', (ebel_relation,), in_out=(variant, basic_node)),
    Edge('has__reactants', (ebel_relation,), in_out=(reactants, basic_node)),
    Edge('has__products', (ebel_relation,), in_out=(products, basic_node)),
    Edge('has__composite', (ebel_relation,), in_out=(composite, basic_node)),

    Edge('has_fragmented_protein', (ebel_relation,), in_out=(bel, bel)),

    has_modified,
    Edge('has_modified_protein', (has_modified,), in_out=(protein, protein)),
    Edge('has_modified_gene', (has_modified,), in_out=(gene, gene)),

    has_variant_obj,
    Edge('has_variant_gene', (has_variant_obj,), in_out=(gene, gene)),
    Edge('has_variant_rna', (has_variant_obj,), in_out=(rna, rna)),
    Edge('has_variant_protein', (has_variant_obj,), in_out=(protein, protein)),
    Edge('has_variant_micro_rna', (has_variant_obj,), in_out=(micro_rna, micro_rna)),

    has_located,
    Edge('has_located_gene', (has_located,), in_out=(gene, gene)),
    Edge('has_located_rna', (has_located,), in_out=(rna, rna)),
    Edge('has_located_protein', (has_located,), in_out=(protein, protein), props=(
        OProperty('levels', ODataType.EMBEDDEDMAP),
    )),
    Edge('has_located_micro_rna', (has_located,), in_out=(micro_rna, micro_rna)),
    Edge('has_located_complex', (has_located,), in_out=(complex_, complex_)),
    Edge('has_located_abundance', (has_located,), in_out=(abundance, abundance)),
    Edge('has_located_population', (has_located,), in_out=(population, population)),
    Edge('pathway_interaction', (ebel_relation,), abstract=True),

    has_ppi,
)

bel_indices = (
    OIndex(bel, ('bel',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel, ('involved_genes',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel, ('involved_other',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel_relation, ('evidence',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(protein, ('uniprot',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(bel_relation, ('annotation',), OIndexType.DICTIONARY),
    OIndex(bel_relation, ('citation',), OIndexType.DICTIONARY),
    OIndex(bel_relation, ('pmid',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ('namespace', 'name'), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ('name',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(namespace_name, ('namespace',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(genetic_flow, ('pure',), OIndexType.NOTUNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of IntAct vertices, edges and indices
##############################################################################
has_ppi_ia = Edge('has_ppi_ia', (has_ppi,), props=(
    OProperty("interaction_type", ODataType.STRING),
    OProperty("interaction_type_psimi_id", ODataType.STRING),
    OProperty("confidence_value", ODataType.FLOAT),
    OProperty("detection_method", ODataType.STRING),
    OProperty("detection_method_psimi_id", ODataType.INTEGER),
    OProperty("pmid", ODataType.INTEGER),
    OProperty("interaction_ids", ODataType.EMBEDDEDMAP),
))

intact_edges: Tuple[Edge, ...] = (
    has_ppi,
    has_ppi_ia,
)

intact_indices = (
    OIndex(has_ppi_ia, ('detection_method', 'interaction_type'), OIndexType.NOTUNIQUE),
)

##############################################################################
# Definition of HGNC generics and indices
# ---------------------------------------
# See .urls.HGNC_JSON
# https://www.genenames.org/help/download
##############################################################################
hgnc = Generic('hgnc', props=(
    OProperty('id', ODataType.STRING),
    OProperty('version', ODataType.INTEGER),
    OProperty('alias_name', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('alias_symbol', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('bioparadigms_slc', ODataType.STRING),
    OProperty('ccds_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('cd', ODataType.STRING),
    OProperty('cosmic', ODataType.STRING),
    OProperty('date_approved_reserved', ODataType.DATE),
    OProperty('date_modified', ODataType.DATE),
    OProperty('date_name_changed', ODataType.DATE),
    OProperty('date_symbol_changed', ODataType.DATE),
    OProperty('ena', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('ensembl_gene_id', ODataType.STRING),
    OProperty('entrez_id', ODataType.INTEGER),
    OProperty('enzyme_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('gene_family', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('gene_family_id', ODataType.EMBEDDEDLIST),
    OProperty('homeodb', ODataType.STRING),
    OProperty('horde_id', ODataType.STRING),
    OProperty('imgt', ODataType.STRING),
    OProperty('intermediate_filament_db', ODataType.STRING),
    OProperty('iuphar', ODataType.STRING),
    OProperty('kznf_gene_catalog', ODataType.INTEGER),
    OProperty('lncipedia', ODataType.STRING),
    OProperty('lncrnadb', ODataType.STRING),
    OProperty('location', ODataType.STRING),
    OProperty('location_sortable', ODataType.STRING),
    OProperty('locus_group', ODataType.STRING),
    OProperty('locus_type', ODataType.STRING),
    OProperty('lsdb', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('mamit_trnadb', ODataType.INTEGER),
    OProperty('merops', ODataType.STRING),
    OProperty('mgd_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('mirbase', ODataType.STRING),
    OProperty('name', ODataType.STRING),
    OProperty('omim_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('orphanet', ODataType.INTEGER),
    OProperty('prev_name', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('prev_symbol', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('pseudogene_org', ODataType.STRING),
    OProperty('pubmed_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.INTEGER),
    OProperty('refseq_accession', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('rgd_id', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('rna_central_ids', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('snornabase', ODataType.STRING),
    OProperty('status', ODataType.STRING),
    OProperty('symbol', ODataType.STRING),
    OProperty('ucsc_id', ODataType.STRING),
    OProperty('uniprot_ids', ODataType.EMBEDDEDLIST, linked_type=ODataType.STRING),
    OProperty('uuid', ODataType.STRING),
    OProperty('vega_id', ODataType.STRING)))

hgnc_generics: Tuple[Generic, ...] = (hgnc,)

hgnc_nodes: Tuple[Node, ...] = (deepcopy(genetic_flow).add_properties([OProperty('hgnc', ODataType.LINK, 'hgnc')]),)

hgnc_indices = (
    OIndex(hgnc, ('symbol',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(hgnc, ('id',), OIndexType.UNIQUE_HASH_INDEX),
    OIndex(hgnc, ('ensembl_gene_id',), OIndexType.NOTUNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of GwasCatalog generics and indices
##############################################################################

mirtarbase_edges: Tuple[Edge, ...] = (
    Edge('has_mirgene_target', (ebel_relation,), props=(
        OProperty('support_type', ODataType.STRING),
        OProperty('pmid', ODataType.INTEGER),
        OProperty('experiments', ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
    ), in_out=(rna, micro_rna)),
)

##############################################################################
# Definition of GwasCatalog generics and indices
##############################################################################
snp = Node('snp', (ebel,), False, (
    OProperty('rs_number', ODataType.STRING, node_view_label=True),
), own_class=False)
gwascatalog_nodes: Tuple[Node, ...] = (snp,)

has_snp = Edge('has_snp', (ebel_relation,), abstract=True, own_class=False, in_out=(snp, gene))
has_mapped_snp = Edge('has_mapped_snp', (has_snp,), abstract=True, own_class=False)
has_downstream_snp = Edge('has_downstream_snp', (has_snp,), abstract=True, own_class=False)
has_upstream_snp = Edge('has_upstream_snp', (has_snp,), abstract=True, own_class=False)
has_snp_gwascatalog = Edge('has_snp_gwascatalog', (has_snp,), abstract=True, props=(
    OProperty('disease_trait', ODataType.STRING),
    OProperty('pubmed_id', ODataType.INTEGER)
))

gwascatalog_edges: Tuple[Edge, ...] = (
    has_snp,
    has_snp_gwascatalog,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge('has_mapped_snp_gc', (has_mapped_snp, has_snp_gwascatalog)),
    Edge('has_downstream_snp_gc', (has_downstream_snp, has_snp_gwascatalog)),
    Edge('has_upstream_snp_gc', (has_upstream_snp, has_snp_gwascatalog)),
)

##############################################################################
# Definition of Ortholog generics and indices
##############################################################################
drug = Node('drug', (ebel,), abstract=True, own_class=False)
drug_db = Node('drug_db', (drug,), props=(
    OProperty('label', ODataType.STRING, node_view_label=True),
    OProperty('drugbank_id', ODataType.STRING),
    OProperty('description', ODataType.STRING),
    OProperty('cas_number', ODataType.STRING, node_view_sub_label=True),
    OProperty('indication', ODataType.STRING),
    OProperty('pharmacodynamics', ODataType.STRING),
    OProperty('toxicity', ODataType.STRING),
    OProperty('metabolism', ODataType.STRING),
    OProperty('mechanism_of_action', ODataType.STRING),
))

drugbank_nodes: Tuple[Node, ...] = (
    drug,
    drug_db,
)

has_drug_target = Edge('has_drug_target', (ebel_relation,), abstract=True, own_class=False)
has_drug_target_db = Edge('has_drug_target_db', (has_drug_target,), in_out=(protein, drug_db), props=(
    OProperty("action", ODataType.STRING),
    OProperty("known_action", ODataType.STRING),
))

drugbank_edges: Tuple[Edge, ...] = (
    has_drug_target,
    has_drug_target_db,
)

drugbank_indices = (
    OIndex(has_drug_target_db, ('action',), OIndexType.NOTUNIQUE_HASH_INDEX),
    OIndex(drug_db, ('drugbank_id',), OIndexType.UNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of IUPHAR's Guide to Pharmacology drugs
##############################################################################

iuphar_interaction = Edge('iuphar_interaction', (ebel_relation,), props=(
    OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
    OProperty("assay_description", ODataType.STRING),
    OProperty("affinity_units", ODataType.STRING),
    OProperty("affinity_low", ODataType.FLOAT),
    OProperty("affinity_median", ODataType.FLOAT),
    OProperty("affinity_high", ODataType.FLOAT),
    OProperty("type", ODataType.STRING),
    OProperty("action", ODataType.STRING),
), in_out=(bel, bel))

iuphar_edges: Tuple[Edge, ...] = (
    iuphar_interaction,
    Edge('agonist_of__iu', (iuphar_interaction,)),
    Edge('inhibits__iu', (iuphar_interaction,)),
    Edge('antagonist_of__iu', (iuphar_interaction,)),
    Edge('channel_blocker_of__iu', (iuphar_interaction,)),
    Edge('allosteric_modulator_of__iu', (iuphar_interaction,)),
    Edge('activates__iu', (iuphar_interaction,)),
    Edge('antibody_against__iu', (iuphar_interaction,)),
    Edge('inhibits_gating__iu', (iuphar_interaction,)),
)

##############################################################################
# Definition of Reactome generics and indices
##############################################################################

reactome_nodes: Tuple[Node, ...] = (
    deepcopy(protein).add_properties(
        [OProperty('reactome_pathways', ODataType.EMBEDDEDSET, linked_type=ODataType.STRING)]),
)

##############################################################################
# Definition of Reactome generics and indices
##############################################################################

has_ppi_bg = Edge('has_ppi_bg', (has_ppi,), abstract=True, props=(
    OProperty("modification", ODataType.STRING),
    OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
    OProperty("dois", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
    OProperty("biogrid_ids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
))

biogrid_edges: Tuple[Edge, ...] = (
    has_ppi_bg,
    Edge('decreases_bg', (has_ppi_bg,)),
)

biogrid_edges_auto_generated = []
for pmod in normalized_pmod.keys():
    for effect in ['increases', 'decreases']:
        edge_name = f"{effect}_{pmod}_bg"
        biogrid_edges_auto_generated.append(Edge(edge_name, (has_ppi_bg,)))

biogrid_edges += tuple(biogrid_edges_auto_generated)

##############################################################################
# Definition of StringDB generics and indices
##############################################################################

has_action = Edge('has_action', (ebel_relation,), abstract=True, own_class=False, in_out=(protein, protein))
has_action_st = Edge('has_action_st', (has_action,), abstract=True)
has_ppi_st = Edge('has_ppi_st', (has_ppi,), abstract=True, props=(
    OProperty("score", ODataType.INTEGER),
))
controls_expression_of_st = Edge('controls_expression_of_st', (has_action_st,), abstract=True)
increases_expression_of = Edge('increases_expression_of', abstract=True, own_class=False)
decreases_expression_of = Edge('decreases_expression_of', abstract=True, own_class=False)

stringdb_edges: Tuple[Edge, ...] = (
    has_ppi,
    Edge('has_ppi_st', (has_ppi,), False, props=(
        OProperty("neighborhood", ODataType.INTEGER),
        OProperty("neighborhood_transferred", ODataType.INTEGER),
        OProperty("fusion", ODataType.INTEGER),
        OProperty("cooccurence", ODataType.INTEGER),
        OProperty("homology", ODataType.INTEGER),
        OProperty("coexpression", ODataType.INTEGER),
        OProperty("coexpression_transferred", ODataType.INTEGER),
        OProperty("experiments", ODataType.INTEGER),
        OProperty("experiments_transferred", ODataType.INTEGER),
        OProperty("database", ODataType.INTEGER),
        OProperty("database_transferred", ODataType.INTEGER),
        OProperty("textmining", ODataType.INTEGER),
        OProperty("textmining_transferred", ODataType.INTEGER),
        OProperty("combined_score", ODataType.INTEGER),
    )),

    has_action,
    has_action_st,
    has_ppi_st,
    Edge('activates_st', (has_action_st,)),
    Edge('inhibits_st', (has_action_st,)),
    controls_expression_of_st,
    increases_expression_of,
    decreases_expression_of,
    Edge('increases_expression_of_st', (has_action_st, controls_expression_of_st, increases_expression_of)),
    Edge('decreases_expression_of_st', (has_action_st, controls_expression_of_st, decreases_expression_of)),
    Edge('controls_pmod_of_st', (has_action_st,)),
)

##############################################################################
# Definition of ClinVar generics and indices
##############################################################################

clinvar_nodes: Tuple[Node, ...] = (
    snp,
)

has_snp_clinvar = Edge('has_snp_clinvar', (has_snp,), abstract=True, props=(
    OProperty('keyword', ODataType.STRING),
    OProperty('clinical_significance', ODataType.STRING),
    OProperty('phenotype', ODataType.STRING),
))

clinvar_edges: Tuple[Edge, ...] = (
    has_snp,
    has_snp_clinvar,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge('has_mapped_snp_cv', (has_mapped_snp, has_snp_clinvar)),
    Edge('has_downstream_snp_cv', (has_downstream_snp, has_snp_clinvar)),
    Edge('has_upstream_snp_cv', (has_upstream_snp, has_snp_clinvar)),
)

clinvar_indices = (
    OIndex(snp, ('rs_number',), OIndexType.UNIQUE_HASH_INDEX),
)

##############################################################################
# Definition of ClinVar generics and indices
# http://stitch.embl.de
##############################################################################
stitch = Generic('stitch', props=(
    OProperty('pubchem_id', ODataType.INTEGER),
    OProperty('ensembl_protein_id', ODataType.INTEGER),
    OProperty('experimental_direct', ODataType.INTEGER),
    OProperty('experimental_transferred', ODataType.INTEGER),
    OProperty('prediction_direct', ODataType.INTEGER),
    OProperty('prediction_transferred', ODataType.INTEGER),
    OProperty('database_direct', ODataType.INTEGER),
    OProperty('database_transferred', ODataType.INTEGER),
    OProperty('textmining_direct', ODataType.INTEGER),
    OProperty('textmining_transferred', ODataType.INTEGER),
    OProperty('combined_score', ODataType.INTEGER),
    OProperty('type', ODataType.STRING),
    OProperty('hgnc', ODataType.LINK, linked_class=hgnc.name),
    OProperty('pubchem', ODataType.LINK, 'stitch_pubchem'),
))

stitch_generics: Tuple[Generic, ...] = (
    Generic('stitch_pubchem', props=(
        OProperty('CID', ODataType.INTEGER),
        OProperty('IUPACName', ODataType.STRING),
        OProperty('MolecularFormula', ODataType.STRING),
        OProperty('IsomericSMILES', ODataType.STRING),
        OProperty('InChI', ODataType.STRING),
        OProperty('InChIKey', ODataType.STRING),
    )),
    stitch,
)

stitch_indices = (
    OIndex(stitch, ("ensembl_protein_id",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("pubchem_id",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("pubchem",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("hgnc",), OIndexType.NOTUNIQUE),
    OIndex(stitch, ("combined_score",), OIndexType.NOTUNIQUE),
)

##############################################################################
# Definition of MirTarBase edges
##############################################################################


##############################################################################
# Definition of DisGeNet
##############################################################################

disgenet_nodes: Tuple[Node, ...] = (
    snp,
)

gene_disease_association = Edge('gene_disease_association', (ebel_relation,), abstract=True)
has_snp_disgenet = Edge('has_snp_disgenet', (has_snp,), abstract=True, props=(
    OProperty('disease_name', ODataType.STRING),
    OProperty('pmid', ODataType.INTEGER),
    OProperty('score', ODataType.FLOAT),
    OProperty('source', ODataType.STRING),
))

disgenet_edges: Tuple[Edge, ...] = (
    gene_disease_association,
    Edge('disgenet_gene_disease', (gene_disease_association,), props=(
        OProperty('pmid', ODataType.INTEGER),
        OProperty('score', ODataType.FLOAT),
        OProperty('source', ODataType.STRING),
    )),

    has_snp,
    has_snp_disgenet,
    has_mapped_snp,
    has_downstream_snp,
    has_upstream_snp,
    Edge('has_mapped_snp_dgn', (has_mapped_snp, has_snp_disgenet)),
    Edge('has_downstream_snp_dgn', (has_downstream_snp, has_snp_disgenet)),
    Edge('has_upstream_snp_dgn', (has_upstream_snp, has_snp_disgenet)),
)

##############################################################################
# Definition of PathwayCommons generics, edges and indices
##############################################################################

pathway_commons_generics: Tuple[Generic, ...] = (
    Generic('pc_pathway_name', props=(
        OProperty('name', ODataType.STRING),
    )),
)
has_action_pc = Edge('has_action_pc', (has_action,), abstract=True, props=(
    OProperty("pathways", ODataType.LINKSET, 'pc_pathway_name'),
    OProperty("sources", ODataType.EMBEDDEDSET, linked_type=ODataType.STRING),
    OProperty("pmids", ODataType.EMBEDDEDSET, linked_type=ODataType.INTEGER),
    OProperty("type", ODataType.STRING),
))

pathway_commons_edges: Tuple[Edge, ...] = (
    has_action,
    has_action_pc,
    Edge('controls_expression_of_pc', (has_action_pc,)),
    Edge('controls_phosphorylation_of_pc', (has_action_pc,)),
    Edge('controls_transport_of_pc', (has_action_pc,)),
)

##############################################################################
# Definition of KEGG generics, edges and indices
##############################################################################

pathway_interaction = Edge('pathway_interaction', (ebel_relation,), abstract=True, own_class=False)
has_ppi_kg = Edge('has_ppi_kg', (pathway_interaction,), abstract=True, props=(
    OProperty("interaction_type", ODataType.STRING),
    OProperty('pathway_names', ODataType.EMBEDDEDSET),
))
increases_expression_of = Edge('increases_expression_of', abstract=True, own_class=False)
decreases_expression_of = Edge('decreases_expression_of', abstract=True, own_class=False)

kegg_edges: Tuple[Edge, ...] = (
    pathway_interaction,
    has_ppi_kg,
    increases_expression_of,
    decreases_expression_of,
    Edge('decreases_pho_kg', (has_ppi_kg,)),
    Edge('increases_gly_kg', (has_ppi_kg,)),
    Edge('increases_me0_kg', (has_ppi_kg,)),
    Edge('increases_pho_kg', (has_ppi_kg,)),
    Edge('increases_ubi_kg', (has_ppi_kg,)),
)

##############################################################################
# NSides definitions
##############################################################################

side_effect = Node('side_effect', (ebel,), props=(
    OProperty('label', ODataType.STRING, node_view_label=True),
    OProperty('condition_meddra_id', ODataType.STRING, node_view_sub_label=True),
))

nsides_nodes: Tuple[Node, ...] = (
    side_effect,
)

nsides_edges: Tuple[Edge] = (
    Edge('has_side_effect', (ebel_relation,), props=(
        OProperty('prr', ODataType.FLOAT),
        OProperty('mean_reporting_frequency', ODataType.FLOAT),
    ), in_out=(side_effect, drug_db)),
)

##############################################################################
# All OrientDB classes as Dict
##############################################################################

nodes: Tuple[Node, ...] = (bel_nodes
                           + hgnc_nodes
                           + gwascatalog_nodes
                           + drugbank_nodes
                           + reactome_nodes
                           + clinvar_nodes
                           + disgenet_nodes
                           + nsides_nodes)

edges: Tuple[Edge, ...] = (bel_edges
                           + intact_edges
                           + mirtarbase_edges
                           + gwascatalog_edges
                           + drugbank_edges
                           + iuphar_edges
                           + biogrid_edges
                           + biogrid_edges
                           + stringdb_edges
                           + clinvar_edges
                           + disgenet_edges
                           + kegg_edges
                           + nsides_edges)

nodes_and_edges: Tuple[OClass, ...] = nodes + edges

nodes_and_edges_dict: Dict[str, OClass] = {o_class.name: o_class for o_class in nodes_and_edges}
non_abstract_nodes_and_edges_dict: Dict[str, OClass] = {o_class.name: o_class for o_class in nodes_and_edges if
                                                        o_class.abstract is False}
abstract_nodes_and_edges_dict: Dict[str, OClass] = {o_class.name: o_class for o_class in nodes_and_edges if
                                                    o_class.abstract is True}


def get_columns(class_name, columns: Tuple[str] = (), exclude_non_serializable: bool = True) -> list:
    """Return list of columns.

    Parameters
    ----------
    class_name : str
        Class name of edge or node.
    columns : List[str], optional
        List of columns.
    exclude_non_serializable : bool
        If True, excludes all non serializable columns.

    Returns
    -------
    list
        List of columns.
    """
    o_class: OClass = nodes_and_edges_dict[class_name]
    exclude_classes = [ODataType.LINK, ODataType.LINKSET, ODataType.LINKMAP, ODataType.LINKBAG]

    columns = list(columns)

    if exclude_non_serializable:
        columns += [p.prop_name for p in o_class.props if p.data_type not in exclude_classes]
    else:
        columns += [p.prop_name for p in o_class.props]
    for extends_name in o_class.extends:
        if extends_name not in ['V', 'E']:
            columns += get_columns(extends_name, tuple(columns), exclude_non_serializable)
    return columns


def get_node_view_labels(name: str, labels: dict = None) -> dict:
    """Get dictionary of labels for a given node.

    Parameters
    ----------
    name : str
        Class name of edge or node.
    labels : dict
        Dictionary of labels and sub labels.

    Returns
    -------
    dict
        {"label": list_of_labels, "sub_label": list_of_sub_labels}
    """
    if not labels:
        labels = {'label': [], 'sub_label': []}
    o_class = nodes_and_edges_dict[name]
    for p in o_class.props:
        if p.node_view_label:
            labels['label'].append(p.prop_name)
        if p.node_view_sub_label:
            labels['sub_label'].append(p.prop_name)
    if not labels['label']:
        for extends_name in o_class.extends:
            if extends_name not in ['V', 'E']:
                get_node_view_labels(extends_name, labels)
    return labels



"""DisGeNet."""
import logging
import pandas as pd

from tqdm import tqdm
from typing import Dict
from pyorientdb import OrientDB

from ebel.manager.orientdb.constants import DISGENET
from ebel.manager.orientdb import odb_meta, urls, odb_structure
from ebel.tools import get_file_path, get_disease_trait_keywords_from_config

from ebel.manager.rdbms.models import disgenet

logger = logging.getLogger(__name__)


class DisGeNet(odb_meta.Graph):
    """DisGeNet (https://www.disgenet.org)."""

    def __init__(self, client: OrientDB = None):
        """Init DisGeNet."""
        self.client = client
        self.biodb_name = DISGENET
        self.urls = {
            'disgenet_gene': urls.DISGENET_GDP_ASSOC,
            'disgenet_variant': urls.DISGENET_VDP_ASSOC,
        }
        super().__init__(tables_base=disgenet.Base,
                         edges=odb_structure.disgenet_edges,
                         urls=self.urls,
                         biodb_name=self.biodb_name)

        self.disease_keywords = get_disease_trait_keywords_from_config()

    def __len__(self):
        return self.number_of_generics

    def __contains__(self, item):
        # TODO: To be implemented
        return True

    def __repr__(self) -> str:
        """Represent DisGeNet as string."""
        template = "{{BioDatabase:DisGeNet}}[url:{url}, edges:{edges}, generics:{generics}]"
        representation = template.format(
            url=self.urls,
            edges=self.number_of_edges,
            generics=self.number_of_generics
        )
        return representation

    def insert_data(self) -> Dict[str, int]:
        """Insert data into database."""
        logger.info(f"Import {self.biodb_name.upper()}")
        inserted = dict()
        inserted['sources'] = self._insert_sources()
        inserted['gene_symbols'] = self._insert_gene_symbols()
        inserted['gene_disease_names'] = self._insert_disease_names()
        inserted['gene_disease_pmid_associations'] = self._insert_gene_disease_pmid_associations()
        inserted['variant_disease_pmid_associations'] = self._insert_variant_disease_pmid_associations()
        return inserted

    def __get_file_for_model(self, model):
        """Return filepath of given model."""
        return get_file_path(self.urls[model.__tablename__], self.biodb_name)

    @property
    def file_path_gene(self):
        """Return filepath of gene."""
        return self.__get_file_for_model(disgenet.DisgenetGene)

    @property
    def file_path_variant(self):
        """Return filepath of variant."""
        return self.__get_file_for_model(disgenet.DisgenetVariant)

    def _insert_sources(self):
        df_g = pd.read_csv(self.file_path_gene, sep="\t", usecols=['source']).drop_duplicates()
        df_v = pd.read_csv(self.file_path_variant, sep="\t", usecols=['source']).drop_duplicates()
        df = pd.concat([df_g, df_v]).drop_duplicates()
        df.reset_index(inplace=True, drop=True)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(disgenet.DisgenetSource.__tablename__, self.engine, if_exists='append')
        return df.shape[0]

    def _insert_disease_names(self) -> int:
        columns_disease = {'diseaseId': 'disease_id', 'diseaseName': 'disease_name'}

        df_gene = pd.read_csv(self.file_path_gene, sep="\t", usecols=columns_disease.keys()) \
            .rename(columns=columns_disease) \
            .drop_duplicates().set_index('disease_id')

        df_variant = pd.read_csv(self.file_path_variant, sep="\t", usecols=columns_disease.keys()) \
            .rename(columns=columns_disease) \
            .drop_duplicates().set_index('disease_id')

        df_concat = pd.concat([df_gene, df_variant]).drop_duplicates()
        df_concat.to_sql(disgenet.DisgenetDisease.__tablename__, self.engine, if_exists='append')
        return df_concat.shape[0]

    def _insert_gene_symbols(self) -> int:
        columns_gene_symols = {'geneId': "gene_id", 'geneSymbol': "gene_symbol"}
        df = pd.read_csv(self.file_path_gene, sep="\t", usecols=columns_gene_symols.keys()) \
            .rename(columns=columns_gene_symols) \
            .drop_duplicates().set_index('gene_id')
        df.to_sql(disgenet.DisgenetGeneSymbol.__tablename__, self.engine, if_exists='append')
        return df.shape[0]

    def _merge_with_source(self, df):
        df_sources = pd.read_sql_table(disgenet.DisgenetSource.__tablename__, self.engine) \
            .rename(columns={'id': 'source_id'})
        return pd.merge(df, df_sources, on="source").drop(columns=['source'])

    def _insert_gene_disease_pmid_associations(self) -> int:
        usecols_gene = ['geneId', 'diseaseId', 'score', 'pmid', 'source']
        rename_dict = dict(zip(usecols_gene, self._standardize_column_names(usecols_gene)))
        df = pd.read_csv(self.file_path_gene, sep="\t", usecols=usecols_gene) \
            .rename(columns=rename_dict)

        df = self._merge_with_source(df)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(disgenet.DisgenetGene.__tablename__, self.engine, if_exists='append')
        return df.shape[0]

    def _insert_variant_disease_pmid_associations(self) -> Dict[str, int]:
        usecols_variant = ['snpId', 'chromosome', 'position', 'diseaseId', 'score', 'pmid', 'source']
        rename_dict = dict(zip(usecols_variant, self._standardize_column_names(usecols_variant)))
        df = pd.read_csv(self.file_path_variant, sep="\t", usecols=usecols_variant) \
            .rename(columns=rename_dict)

        df = self._merge_with_source(df)
        df.index += 1
        df.index.rename('id', inplace=True)
        df.to_sql(disgenet.DisgenetVariant.__tablename__, self.engine, if_exists='append')

        return df.shape[0]

    def update_interactions(self) -> int:
        """Update edges with disgenet metadata."""
        self.clear_nodes_and_edges()
        # self.update_diseases()
        inserted = self.update_snps()
        self.delete_nodes_with_no_edges('snp')
        return inserted

    def update_diseases(self):
        """Update disease information."""
        # TODO: to be implemented, but disease node class did not exist before and definitions are unclear
        # query = self.session.query(disgenet.DisgenetDisease).filter(
        #     disgenet.DisgenetDisease.disease_name.like('%alzheimer%'))
        # diseases = [x.disease_id for x in query.all()]
        pass

    def update_snps(self) -> int:
        """Update SNP information."""
        snp_type = {'mapped': "mapped", 'downstream': "upstream", 'upstream': "downstream"}
        # TODO: replace SQL with SQL Alchemy statement
        sql_temp = """Select
                snp_id,
                chromosome,
                position,
                disease_name,
                pmid,
                score,
                source
            FROM
                disgenet_variant v INNER JOIN
                disgenet_source s on (v.source_id=s.id) INNER JOIN
                disgenet_disease d on (v.disease_id=d.disease_id)
            WHERE
                disease_name like '%%{}%%' and
                source!='BEFREE'
            GROUP BY
                snp_id,
                chromosome,
                position,
                disease_name,
                pmid,
                score,
                source"""

        results = dict()
        for kwd in self.disease_keywords:
            sql = sql_temp.format(kwd)
            rows = self.engine.execute(sql)
            results[kwd] = rows

        inserted = 0

        snps = {x['rs_number']: x['rid'] for x in self.query_class('snp', columns=['rs_number'])}

        for trait, kwd_disease_results in results.items():
            for r in tqdm(kwd_disease_results,
                          desc=f'Update DisGeNET variant interactions for {trait}',
                          total=kwd_disease_results.rowcount):
                snp_id, chromosome, position, disease_name, pmid, score, source = r

                if snp_id in snps:
                    snp_rid = snps[snp_id]

                else:
                    snp_rid = self.insert_record(class_name='snp', value_dict={'rs_number': snp_id})
                    snps[snp_id] = snp_rid

                gene_type_rids = self.get_set_gene_rids_by_position(chromosome, position)

                for gene_type, gene_rids in gene_type_rids.items():
                    for gene_rid in gene_rids:
                        value_dict = {'disease_name': disease_name,
                                      'score': score,
                                      'source': source,
                                      'pmid': pmid
                                      }
                        class_name = f"has_{snp_type[gene_type]}_snp_dgn"
                        self.create_edge(class_name=class_name,
                                         from_rid=gene_rid,
                                         to_rid=snp_rid,
                                         value_dict=value_dict)
                inserted += 1

        return inserted



# -*- coding: utf-8 -*-

"""This file contains default values for configurations and parameters."""

import os
import logging
import logging.handlers as handlers

from .constants import PROJECT_DIR, DATA_DIR, LOG_DIR

###############################################################################
# UNIPROT taxonomy IDs to import

default_tax_ids = [
    9606,  # Humans
    10116,  # Rats
    10090,  # Mice
    7227,  # Drosophila melanogaster
    694009, 2697049,  # COVID
    7955,  # Zebrafish
]
###############################################################################

SQLITE_DATABASE_NAME = 'ebel.db'
SQLITE_TEST_DATABASE_NAME = 'ebel_test.db'
DATABASE_LOCATION = os.path.join(
    DATA_DIR,
    SQLITE_DATABASE_NAME
)
DEFAULT_TEST_DATABASE_LOCATION = os.path.join(
    DATA_DIR,
    SQLITE_TEST_DATABASE_NAME
)

###############################################################################
# SQLAlchemy connection strings
# =============================
# SQLite
# ------
CONN_STR_DEFAULT = 'sqlite:///' + DATABASE_LOCATION
CONN_STR_TESTS = 'sqlite:///' + SQLITE_TEST_DATABASE_NAME
# MySQL
# -----
CONN_STR_MYSQL_PREFIX = 'mysql+pymysql://ebel:ebel@localhost/'
CONN_STR_MYSQL = CONN_STR_MYSQL_PREFIX + 'ebel?charset=utf8'
CONN_STR_MYSQL_TESTS = CONN_STR_MYSQL_PREFIX + 'ebel_test?charset=utf8'

###############################################################################
# Config
config_file_path = os.path.join(PROJECT_DIR, 'config.ini')

###############################################################################
# Log Handling
logHandler = handlers.RotatingFileHandler(filename=os.path.join(LOG_DIR, 'ebel.log'),
                                          mode='a',
                                          maxBytes=4098 * 10,  # 4MB file max
                                          backupCount=0)
logh_format = logging.Formatter('%(asctime)s - %(name)s - %(levelname)s - %(message)s')
logHandler.setFormatter(logh_format)
logHandler.setLevel(logging.DEBUG)

# Console Handler
ch = logging.StreamHandler()
ch_format = logging.Formatter('%(asctime)s - %(levelname)s - %(message)s')
ch.setFormatter(ch_format)
ch.setLevel(logging.WARNING)



"""nSIDES API methods."""

from flask import request

from ebel.manager.rdbms.models.nsides import Nsides
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_ebel_query_result


def get_nsides():
    """Get generic NSIDES entry."""
    return _get_data(Nsides)


def get_ebel_relation():
    """Get NSIDES related eBEL relations."""
    conf = {
        'drugbank_id': "out.drugbank_id",
        'drug_label': "out.label",
        'ppr': "prr",
        'mean_reporting_frequency': "mean_reporting_frequency",
        'condition_meddra_id': "in.condition_meddra_id",
        'side_effect': "in.label",
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_side_effect"

    paras = {k: request.args[k] for k in request.args if k in conf}
    if paras:
        wheres = [f'{conf[k].replace(".asString()","")} like "{v}"' for k, v in paras.items()]
        sql += " WHERE " + ' AND '.join(wheres)

    return _get_paginated_ebel_query_result(sql)



"""NCBI RDBMS model definition."""
from sqlalchemy.orm import relationship
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy import Column, Integer, String, Text, ForeignKey

from . import object_as_dict

Base = declarative_base()


class NcbiGeneInfo(Base):
    """Class definition for the ncbi_gene_info table."""

    __tablename__ = 'ncbi_gene_info'
    gene_id = Column(Integer, primary_key=True)

    tax_id = Column(Integer, index=True)
    symbol = Column(String(100), index=True)
    type_of_gene = Column(String(100), index=True)
    locus_tag = Column(String(100))
    chromosome = Column(String(100))
    map_location = Column(String(100))
    description_id = Column(Integer, ForeignKey('ncbi_gene_info_description.id'))
    description = relationship("NcbiGeneInfoDescription", foreign_keys=[description_id])
    xrefs = relationship("NcbiGeneInfoXref", back_populates="gene")
    mims = relationship("NcbiGeneMim", foreign_keys='NcbiGeneMim.gene_id', back_populates="gene")
    orthologs = relationship("NcbiGeneOrtholog", foreign_keys='NcbiGeneOrtholog.gene_id', back_populates="gene")
    ensembl_ids = relationship("NcbiGeneEnsembl", back_populates="genes")
    gene_ids_right = relationship("NcbiGeneOnRight", foreign_keys='NcbiGeneOnRight.gene_id', back_populates="gene")
    gene_ids_left = relationship("NcbiGeneOnLeft", foreign_keys='NcbiGeneOnLeft.gene_id', back_populates="gene")
    gene_ids_overlapping = relationship("NcbiGeneOverlapping", foreign_keys='NcbiGeneOverlapping.gene_id',
                                        back_populates="gene")

    def as_dict(self):
        """Convert object values to dictionary."""
        rdict = object_as_dict(self, ['description_id'])
        rdict.update({
            'xrefs': [{'db': x.db, 'dbid': x.dbid} for x in self.xrefs],
            'mims': [x.mim_number for x in self.mims],
            'description': self.description.description,
            'ensembls': [x.ensembl_gene_identifier for x in self.ensembl_ids],
            'orthologs': [x.other_gene_id for x in self.orthologs],
            'gene_ids_right': [x.gene_id_on_right for x in self.gene_ids_right],
            'gene_ids_left': [x.gene_id_on_left for x in self.gene_ids_left],
            'gene_ids_overlapping': [x.overlapping_gene_id for x in self.gene_ids_overlapping]
        })
        return rdict


class NcbiGeneInfoDescription(Base):
    """Class definition for the ncbi_gene_info_description table."""

    __tablename__ = 'ncbi_gene_info_description'
    id = Column(Integer, primary_key=True, autoincrement=True)
    description = Column(Text)


class NcbiGeneOnRight(Base):
    """Class definition for the ncbi_gene_on_right table."""

    __tablename__ = 'ncbi_gene_on_right'
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    gene_id_on_right = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOnLeft(Base):
    """Class definition for the ncbi_gene_on_left table."""

    __tablename__ = 'ncbi_gene_on_left'
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    gene_id_on_left = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOverlapping(Base):
    """Class definition for the ncbi_gene_overlapping table."""

    __tablename__ = 'ncbi_gene_overlapping'
    id = Column(Integer, primary_key=True, autoincrement=True)
    gene_id = gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    overlapping_gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGeneOrtholog(Base):
    """Class definition for the ncbi_gene_ortholog table."""

    __tablename__ = 'ncbi_gene_ortholog'
    id = Column(Integer, primary_key=True, autoincrement=True)
    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    other_tax_id = Column(Integer, index=True)
    other_gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))

    gene = relationship("NcbiGeneInfo", foreign_keys=[gene_id])


class NcbiGenePubmed(Base):
    """Class definition for the ncbi_gene_pubmed table."""

    __tablename__ = 'ncbi_gene_pubmed'
    id = Column(Integer, primary_key=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    pub_med_id = Column(Integer)


class NcbiGeneInfoXref(Base):
    """Class definition for the ncbi_gene_info_xref table."""

    __tablename__ = "ncbi_gene_info_xref"
    id = Column(Integer, primary_key=True, autoincrement=True)

    db = Column(String(100), index=True)
    dbid = Column(String(100), index=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))

    gene = relationship("NcbiGeneInfo", back_populates="xrefs")


class NcbiGeneMim(Base):
    """Class definition for the ncbi_gene_mim table."""

    __tablename__ = 'ncbi_gene_mim'
    id = Column(Integer, primary_key=True, autoincrement=True)

    mim_number = Column(Integer)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    type = Column(String(100))
    source = Column(String(100))
    med_gen_cui = Column(String(100), index=True)
    comment = Column(String(100))

    gene = relationship("NcbiGeneInfo", back_populates="mims")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "mim_number": self.mim_number,
            "gene_id": self.gene_id,
            "type": self.type,
            "source": self.source,
            "med_gen_cui": self.med_gen_cui,
            "comment": self.comment
        }


class NcbiGeneEnsembl(Base):
    """Class definition for the ncbi_gene_ensembl table."""

    __tablename__ = 'ncbi_gene_ensembl'
    id = Column(Integer, primary_key=True, autoincrement=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    ensembl_gene_identifier = Column(String(100))
    rna_nucleotide_accession_version = Column(String(100))
    ensembl_rna_identifier = Column(String(100))
    protein_accession_version = Column(String(100))
    ensembl_protein_identifier = Column(String(100))

    genes = relationship("NcbiGeneInfo", back_populates="ensembl_ids")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            "tax_id": self.tax_id,
            "gene_id": self.gene_id,
            "ensembl_gene_identifier": self.ensembl_gene_identifier,
            "rna_nucleotide_accession_version": self.rna_nucleotide_accession_version,
            "ensembl_rna_identifier": self.ensembl_rna_identifier,
            "protein_accession_version": self.protein_accession_version,
            "ensembl_protein_identifier": self.ensembl_protein_identifier
        }


class NcbiGeneGo(Base):
    """Class definition for the ncbi_gene_go table."""

    __tablename__ = 'ncbi_gene_go'
    id = Column(Integer, primary_key=True, autoincrement=True)

    tax_id = Column(Integer, index=True)
    gene_id = Column(Integer, ForeignKey('ncbi_gene_info.gene_id'))
    go_id = Column(String(100), index=True)
    evidence = Column(String(10))
    qualifier = Column(String(100))
    go_term = Column(String(255))
    category = Column(String(10))

    pmids = relationship("NcbiGeneGoPmid", back_populates="gos")

    def as_dict(self):
        """Convert object values to dictionary."""
        return {
            'tax_id': self.tax_id,
            'gene_id': self.gene_id,
            'go_id': self.go_id,
            'evidence': self.evidence,
            'qualifier': self.qualifier,
            'go_term': self.go_term,
            'pmids': [x.pmid for x in self.pmids],
            'category': self.category
        }


class NcbiGeneGoPmid(Base):
    """Class definition for the ncbi_gene_go_pmid table."""

    __tablename__ = 'ncbi_gene_go_pmid'
    id = Column(Integer, primary_key=True, autoincrement=True)

    ncbi_gene_go_id = Column(Integer, ForeignKey('ncbi_gene_go.id'))
    pmid = Column(Integer)

    gos = relationship("NcbiGeneGo", back_populates="pmids")


class NcbiMedGenName(Base):
    """Class definition for the ncbi_medgen_name table."""

    __tablename__ = 'ncbi_medgen_name'
    id = Column(Integer, primary_key=True, autoincrement=True)

    cui = Column(String(100))
    name = Column(Text)
    source = Column(String(100))
    suppress = Column(String(1))
    pmids = relationship("NcbiMedGenPmid", back_populates="med_gen_name")

    def as_dict(self):
        """Convert object values to dictionary."""
        rdit = object_as_dict(self, exclude=['id'])
        rdit.update({'pmids': [x.pmid for x in self.pmids]})
        return rdit


class NcbiMedGenPmid(Base):
    """Class definition for the ncbi_medgen_pmid table."""

    __tablename__ = 'ncbi_medgen_pmid'
    id = Column(Integer, primary_key=True, autoincrement=True)

    ncbi_medgen_name_id = Column(Integer, ForeignKey('ncbi_medgen_name.id'))
    pmid = Column(Integer, index=True)

    med_gen_name = relationship("NcbiMedGenName", back_populates="pmids")

    def as_dict(self):
        """Convert object values to dictionary."""
        rdit = object_as_dict(self, exclude=['id'])
        return rdit



"""Pathway Commons API unit tests."""

from .conftest import format_response_data
from .constants import RESULTS, NUM_RESULTS, PAGE_SIZE


class TestPathwayCommons:
    def test_get_pathway_commons(self, client):
        response = client.get(
            'api/v1/pathway_commons?participant_a=CSF3&interaction_type=controls-expression-of&participant_b=CD33&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_results = {
          "id": 804357,
          "interaction_type": "controls-expression-of",
          "participant_a": "CSF3",
          "participant_b": "CD33",
          "pathway_names": [],
          "pmids": [
            12627849
          ],
          "sources": [
            "CTD"
          ]
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_results

    def test_get_by_gene_symbol(self, client):
        response = client.get(
            'api/v1/pathway_commons/by_gene_symbol?gene_symbol=CD33&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_cols = ("id", "interaction_type", "participant_a", "participant_b", "pathway_names", "pmids",
                         "sources")
        assert output[NUM_RESULTS] >= 19
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert hit["participant_b"] == "CD33"

    def test_get_by_pathway_name(self, client):
        response = client.get(
            'api/v1/pathway_commons/by_pathway_name?pathway_name=Activation%20and%20oligomerization%20of%20BAK%20protein&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_result = {
          "id": 69305,
          "interaction_type": "in-complex-with",
          "participant_a": "BAK1",
          "participant_b": "BID",
          "pathway_names": [
            "Activation and oligomerization of BAK protein"
          ],
          "pmids": [],
          "sources": [
            "Reactome"
          ]
        }
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert hit == expected_result

    def test_get_by_pmid(self, client):
        response = client.get(
            'api/v1/pathway_commons/by_pmid?pmid=12627849&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        expected_cols = ("id", "interaction_type", "participant_a", "participant_b", "pathway_names", "pmids",
                         "sources")
        assert output[NUM_RESULTS] >= 6
        assert output[PAGE_SIZE] == 10
        results = output[RESULTS]
        assert isinstance(results, list)
        hit = results[0]
        assert isinstance(hit, dict)
        assert all([col in hit for col in expected_cols])
        assert 12627849 in hit["pmids"]

    def test_get_pathway_name_starts_with(self, client):
        response = client.get(
            'api/v1/pathway_commons/pathway_name/starts_with?pathway_name=Activation%20and%20oligomerization%20of%20BAK&page_size=10&page=1',
            content_type='application/json'
        )
        output = format_response_data(response)
        assert output[NUM_RESULTS] == 1
        assert output[PAGE_SIZE] == 10
        hit = output[RESULTS]

        assert isinstance(hit, dict)
        assert "Activation and oligomerization of BAK protein" in hit
        assert isinstance(hit["Activation and oligomerization of BAK protein"], int)



"""DrugBank API methods."""
from flask import request
from sqlalchemy.orm.attributes import InstrumentedAttribute

from ebel.web.api import RDBMS
from ebel.manager.rdbms.models import drugbank
from ebel.web.api.ebel.v1 import _get_data, _get_paginated_query_result, _get_paginated_ebel_query_result


def get_by_id():
    """Get DrugBank entry by ID."""
    drugbank_id = request.args.get('drugbank_id')
    if drugbank_id:
        query = RDBMS.get_session().query(drugbank.Drugbank).filter_by(drugbank_id=drugbank_id.strip())
        return query.first().as_dict()


def get_drugbank():
    """Get generic DrugBank entry."""
    return _get_data(drugbank.Drugbank)


def get_interaction():
    """Get DrugBank drug interactions."""
    return _get_data(drugbank.DrugInteraction)


def _get_model(model):
    """Get tables connected to Drugbank - paginated."""
    drugbank_id = request.args.get('drugbank_id')
    columns_like = {
        col_obj.like(request.args[col_name]) for col_name, col_obj in model.__dict__.items()
        if isinstance(col_obj, InstrumentedAttribute) and col_name in request.args and col_name != 'drugbank_id'
    }
    query = RDBMS.get_session().query(model).filter(*columns_like)
    if drugbank_id:
        query = query.join(drugbank.Drugbank).filter_by(drugbank_id=drugbank_id)
    return _get_paginated_query_result(query, print_sql=True)


def get_pathway():
    """Get DrugBank involved pathways."""
    return _get_model(drugbank.Pathway)


def get_status():
    """Get DrugBank statuses."""
    return _get_model(drugbank.Status)


def get_patent():
    """Get DrugBank patents."""
    return _get_model(drugbank.Patent)


def get_external_identifier():
    """Get DrugBank external identifiers."""
    return _get_model(drugbank.ExternalIdentifier)


def get_reference():
    """Get DrugBank references."""
    return _get_model(drugbank.Reference)


def get_target():
    """Get DrugBank targets."""
    return _get_model(drugbank.Target)


def get_product_name():
    """Get DrugBank product names."""
    return _get_model(drugbank.ProductName)


def get_synonym():
    """Get DrugBank synonyms."""
    return _get_model(drugbank.Synonym)


def get_has_drug_target_db():
    """Get DrugBank related eBEL relations."""
    conf = {
        'target_rid': "in.@rid.asString()",
        'target_name': "in.name",
        'target_namespace': "in.namespace",
        'target_uniprot_accession': "in.uniprot",
        'drugbank_id': "out.drugbank_id",
        'drug_name': "out.label",
        'drug_rid': "out.@rid.asString()",
        'edge_rid': "@rid.asString()",
        'action': "action",
        'known_action': "known_action"
    }
    sql = "SELECT "
    sql += ', '.join([f"{v} as {k}" for k, v in conf.items()])
    sql += " FROM has_drug_target_db"
    ra = request.args
    paras = {k: ra[k] for k in ra if k in conf}
    if paras:
        sql += " WHERE " + ' AND '.join([f'{conf[k]} = "{v}"' for k, v in paras.items()])
    return _get_paginated_ebel_query_result(sql)
